#!/usr/bin/env python
import os

import fastr
from fastr.utils.rest_generation import create_rest_table


def generate_config(fields):
    field_names = []
    field_types = []
    field_defaults = []
    field_description = []

    for name, field in sorted(fields.items()):
        field_names.append(name)
        field_types.append(field[0].__name__)
        field_description.append(field[2])

        if len(field) > 3:
            field_defaults.append(field[3] if isinstance(field[3], basestring) else repr(field[3]))
        else:
            field_defaults.append(repr(field[1]))

    data = [field_names, field_types, field_description, field_defaults]
    headers = ['name', 'type', 'description', 'default']

    return create_rest_table(data, headers)


def generate_config_doc():
    filename = os.path.join(os.path.dirname(__file__), 'fastr.config.rst')

    with open(filename, 'w') as fh_out:
        fh_out.write(generate_config(fastr.config.DEFAULT_FIELDS))


if __name__ == '__main__':
    generate_config_doc()



import argparse


def main():
    parser = argparse.ArgumentParser(description='Short sample app')
    parser.add_argument('--in_1', type=int)
    parser.add_argument('--in_2', type=int)
    parser.add_argument('--fail_1', action="store_true", default=False)
    parser.add_argument('--fail_2', action="store_true", default=False)

    args = parser.parse_args()

    print(args)
    print('in 1  : {}'.format(args.in_1))
    print('in 2  : {}'.format(args.in_2))
    print('fail_1: {}'.format(args.fail_1))
    print('fail_2: {}'.format(args.fail_2))

    if not args.fail_1:
        print('RESULT_1=[{}]'.format(args.in_1 + 1))
    if not args.fail_2:
        print('RESULT_2=[{}]'.format(args.in_2 + 1))


if __name__ == "__main__":
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the FileSystem plugin for fastr
"""

import os
import shutil
import urlparse as up
import fastr
import fastr.exceptions as exceptions
from fastr.core.ioplugin import IOPlugin


class FileSystem(IOPlugin):
    """
    The FileSystem plugin is create to handle ``file://`` type or URLs. This is
    generally not a good practice, as this is not portable over between
    machines. However, for test purposes it might be useful.

    The URL scheme is rather simple: ``file://host/path``
    (see `wikipedia <http://en.wikipedia.org/wiki/File_URI_scheme>`_ for details)

    We do not make use of the ``host`` part and at the moment only support
    localhost (just leave the host empty) leading to ``file:///`` URLs.

    .. warning:: This plugin ignores the hostname in the URL and does only
                 accept driver letters on Windows in the form ``c:/``
    """
    scheme = 'file'

    def __init__(self):
        # initialize the instance and register the scheme
        super(FileSystem, self).__init__()

    def url_to_path(self, url):
        """ Get the path to a file from a url.
        Currently supports the file:// scheme

        Examples:

        .. code-block:: python

          >>> 'file:///d:/data/project/file.ext'
          'd:\\data\\project\\file.ext'

        .. warning::

          file:// will not function cross platform and is mainly for testing

        """
        parsed_url = up.urlparse(str(url))

        # Translate properly depending on the scheme being used
        if parsed_url.scheme == self.scheme:
            if os.name == 'nt':
                path = parsed_url.path.lstrip('/')
            else:
                path = parsed_url.path

            return self._correct_separators(path)
            # return o.path
        else:
            raise exceptions.FastrValueError('This parses the {} scheme and not the {} scheme!'.format(self.scheme,
                                                                                                       parsed_url.scheme))

    def path_to_url(self, path, mountpoint=None):
        """ Construct an url from a given mount point and a relative path to the mount point. """
        return "{scheme}:///{path}".format(scheme=self.scheme, path=path)

    def fetch_url(self, inurl, outpath):
        """
        Fetch the files from the file.

        :param inurl: url to the item in the data store, starts with ``file://``
        :param outpath: path where to store the fetch data locally
        """
        inpath = self.url_to_path(inurl)

        # Clear away present data
        if os.path.exists(outpath):
            fastr.log.info('Removing currently exists data at {}'.format(outpath))
            if os.path.islink(outpath):
                os.remove(outpath)
            elif os.path.isdir(outpath):
                shutil.rmtree(outpath)
            else:
                os.remove(outpath)

        try:
            os.symlink(inpath, outpath)
            fastr.log.debug('Symlink successful')
        except OSError:
            fastr.log.debug('Cannot symlink, fallback to copy')
            if os.path.isdir(inpath):
                shutil.copytree(inpath, outpath)
            else:
                shutil.copy2(inpath, outpath)

        return outpath

    def fetch_value(self, inurl):
        """
        Fetch a value from an external file file.

        :param inurl: url of the value to read
        :return: the fetched value
        """
        path = self.url_to_path(inurl)
        with open(path, 'r') as file_handle:
            data = file_handle.read()

        return data

    def put_url(self, inpath, outurl):
        """
        Put the files to the external data store.

        :param inpath: path of the local data
        :param outurl: url to where to store the data, starts with ``file://``
        """
        outpath = self.url_to_path(outurl)
        fastr.vfs.copy_file_dir(inpath, outpath)
        return os.path.exists(outpath)

    def put_value(self, value, outurl):
        """
        Put the value in the external data store.

        :param value: value to store
        :param outurl: url to where to store the data, starts with ``file://``
        """
        outpath = self.url_to_path(outurl)

        with open(outpath, 'w') as file_handle:
            file_handle.write(str(value))

        return os.path.exists(outpath)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Job class and some related classes.
"""

from collections import OrderedDict
import datetime
import os
import re
import urllib
import urlparse

from enum import Enum

import fastr
from fastr.core.inputoutput import Output
from fastr.core.provenance import Provenance
from fastr.core.samples import SampleItem
from fastr.core.serializable import Serializable
from fastr.data import url
from fastr.datatypes import URLType, DataType, Deferred
import fastr.exceptions as exceptions
from fastr.utils.iohelpers import load_gpickle, save_gpickle, save_json


try:
    from fastr.execution.environmentmodules import EnvironmentModules
    ENVIRONMENT_MODULES = EnvironmentModules(fastr.config.protected_modules)
    ENVIRONMENT_MODULES_LOADED = True
except exceptions.FastrValueError:
    ENVIRONMENT_MODULES_LOADED = False


class JobState(Enum):
    """
    The possible states a Job can be in. An overview of the states and the
    adviced transitions are depicted in the following figure:

    .. graphviz::

       digraph jobstate {
           nonexistent [shape=box];
           created [shape=box];
           queued [shape=box];
           hold [shape=box];
           running [shape=box];
           execution_done [shape=box];
           execution_failed [shape=box];
           processing_callback [shape=box];
           finished [shape=box];
           failed [shape=box];
           cancelled [shape=box];

           nonexistent -> created;
           created -> queued;
           created -> hold;
           hold -> queued;
           queued -> running;
           running -> execution_done;
           running -> execution_failed;
           execution_done -> processing_callback;
           execution_failed -> processing_callback;
           processing_callback -> finished;
           processing_callback -> failed;
           running -> cancelled;
           queued -> cancelled;
           hold -> cancelled;
       }
    """
    nonexistent = ('nonexistent', 'idle', False)
    created = ('created', 'idle', False)
    queued = ('queued', 'idle', False)
    hold = ('hold', 'idle', False)
    running = ('running', 'in_progress', False)
    execution_done = ('execution_done', 'in_progress', False)
    execution_failed = ('execution_failed', 'in_progress', True)
    processing_callback = ('processing_callback', 'in_progress', False)
    finished = ('finished', 'done', False)
    failed = ('failed', 'done', True)
    cancelled = ('cancelled', 'done', True)

    def __init__(self, _, stage, error):
        self.stage = stage
        self.error = error

    @property
    def done(self):
        return self.stage == 'done'

    @property
    def in_progress(self):
        return self.stage == 'in_progress'


class Job(Serializable):

    """Class describing a job.

       Arguments:
       tool_name - the name of the tool (str)
       tool_version - the version of the tool (Version)
       argument - the arguments used when calling the tool (list)
       tmpdir - temporary directory to use to store output data
       hold_jobs - list of jobs that need to finished before this job can run (list)
    """

    # Constants for file names related to jobs
    COMMAND_DUMP = '__fastr_command__.pickle.gz'
    RESULT_DUMP = '__fastr_result__.pickle.gz'
    STDOUT_DUMP = '__fastr_stdout__.txt'
    STDERR_DUMP = '__fastr_stderr__.txt'
    INFO_DUMP = '__fastr_extra_job_info__.json'
    PROV_DUMP = '__fastr_prov__.json'

    def __init__(self, node, sample_id, sample_index,
                 input_arguments, output_arguments, hold_jobs=None,
                 status_callback=None, preferred_types=None):
        """
        Create a job

        :param Node node: the node the job is based on
        :param fastr.core.samples.SampleId sample_id: the id of the sample
        :param fastr.core.samples.SampleIndex sample_index: the index of the sample
        :param list[dict] input_arguments: the argument list
        :param list[dict] output_arguments: the argument list
        :param list[str] hold_jobs: the jobs on which this jobs depend
        :param callable status_callback: The callback to call when the status changed
        :param preferred_types: The list of preferred types to use
        :return:
        """
        # Save information about the Job environment in which it was created
        self.network_id = node.parent.id
        self.network_version = str(node.parent.version)
        self.run_id = node.parent.run_id
        self.node_id = node.id
        self.tool_name = node.tool.ns_id
        self.tool_version = str(node.tool.command['version'])
        self.sample_id = sample_id
        self.sample_index = sample_index
        self.network_tmpurl = node.parent.tmpurl

        # Arguments for the command
        self.input_arguments = input_arguments
        self.output_arguments = output_arguments

        # Create tmpdir
        # Determine subfolder name in tmp mount
        self.ensure_tmp_dir()

        self._required_cores = None
        self._required_memory = None
        self._required_time = None
        self.required_cores = node.required_cores
        self.required_memory = node.required_memory
        self.required_time = node.required_time
        self.translated_values = {}
        self.status_callback = status_callback
        self.preferred_types = preferred_types if preferred_types else {}

        if isinstance(hold_jobs, (set, list, tuple)):
            self.hold_jobs = list(hold_jobs)
        elif isinstance(hold_jobs, str):
            self.hold_jobs = [hold_jobs]
        elif hold_jobs is None:
            self.hold_jobs = []
        else:
            raise exceptions.FastrTypeError('Cannot create jobs: hold_jobs has invalid type!')

        self.timestamp = datetime.datetime.utcnow()

        self.info_store = {
            'id': self.id,
        }

        self.errors = []
        self.input_hash = {}
        self.output_hash = {}

        # The status list contains tuples with (timestamp, status)
        # This is updated via the status property
        self.status_list = []
        self.status = JobState.created

        self.provenance = Provenance()

        # Dictionary where the output data will be stored
        self.output_data = {}

        # Save fastr version for safety
        self.fastr_version = fastr.version.full_version

    def __getstate__(self):
        """
        Get the state of the job

        :return: job state
        :rtype: dict
        """
        state = {k: v for k, v in self.__dict__.items()}
        del state['status_callback']
        del state['info_store']
        del state['provenance']
        return state

    def __setstate__(self, state):
        """
        Set the state of the job

        :param dict state:
        """
        # For backwards compatibility
        if 'status_list' not in state and '_status_list' in state:
            state['status_list'] = state.pop('_status_list')

        self.status_callback = None
        self.__dict__.update(state)

        if not hasattr(self, 'info_store'):
            self.info_store = {
                'id': self.id,
            }

        if not hasattr(self, 'provenance'):
            self.provenance = Provenance()

    def get_result(self):
        """
        Get the result of the job if it is available. Load the output file if
        found and check if the job matches the current object. If so, load and
        return the result.

        :returns: Job after execution or None if not available
        :rtype: Job | None
        """

        if not os.path.exists(self.logfile):
            return None

        fastr.log.debug('Found old job result file: {}'.format(self.logfile))

        try:
            result = load_gpickle(self.logfile)
        except (IOError, EOFError):
            # Errors loading pickle or gzip stream
            return None

        fastr.log.debug('Loaded old job result file')

        if not isinstance(result, Job):
            fastr.log.debug('Result is not valid Job! (found {})'.format(type(result).__name__))
            return None

        if result.status != JobState.execution_done:
            fastr.log.debug('Result status is wrong ({})'.format(result.status))
            return None

        if result.id != self.id:
            fastr.log.debug('Result job id is wrong ({})'.format(result.id))
            return None

        if result.tool_name != self.tool_name:
            fastr.log.debug('Result tool name is wrong ({})'.format(result.tool_name))
            return None

        if result.tool_version != self.tool_version:
            fastr.log.debug('Result tool version is wrong ({})'.format(result.tool_version))
            return None

        if result.sample_id != self.sample_id:
            fastr.log.debug('Result sample id is wrong ({})'.format(result.sample_id))
            return None

        fastr.log.debug('Checking payloads')

        result_payload = result.create_payload()
        if result_payload != self.create_payload():
            fastr.log.debug('Result payload is wrong ({})'.format(result_payload))
            return None

        fastr.log.debug('Checking sample index')

        if self.sample_index != result.sample_index:
            fastr.log.info('Updating sample index from {} to {}'.format(result.sample_index,
                                                                        self.sample_index))
            result.sample_index = self.sample_index

        return result

    def __repr__(self):
        """
        String representation of the Job
        """
        return '<Job\n  id={job.id}\n  tool={job.tool_name} {job.tool_version}\n  tmpdir={job.tmpurl}/>'.format(job=self)

    @property
    def status(self):
        """
        The status of the job
        """
        # The get last status update and only the new status
        return self.status_list[-1][1]

    @status.setter
    def status(self, status):
        """
        Set the status of a job
        :param status: new status
        """
        if not isinstance(status, JobState):
            raise exceptions.FastrTypeError('Job status should be of class JobState, found [{}] {}'.format(type(status).__name__,
                                                                                                           status))
        if len(self.status_list) == 0 or self.status != status:
            self.status_list.append((datetime.datetime.utcnow(), status))

            if self.status_callback is not None:
                self.status_callback(self)

    @property
    def id(self):
        """
        The id of this job
        """
        return '{}___{}___{}'.format(self.network_id,
                                     self.node_id,
                                     self.sample_id)

    @property
    def fullid(self):
        """
        The full id of the job
        """
        return self.id
    
    @property
    def tmpurl(self):
        """
        The URL of the tmpdir to use
        """
        return url.join(self.network_tmpurl, self.node_id, str(self.sample_id))

    @property
    def commandurl(self):
        """
        The url of the command pickle
        """
        return url.join(self.tmpurl, self.COMMAND_DUMP)

    @property
    def logurl(self):
        """
        The url of the result pickle
        """
        return url.join(self.tmpurl, self.RESULT_DUMP)

    @property
    def stdouturl(self):
        """
        The url where the stdout text is saved
        """
        return url.join(self.tmpurl, self.STDOUT_DUMP)

    @property
    def stderrurl(self):
        """
        The url where the stderr text is saved
        """
        return url.join(self.tmpurl, self.STDERR_DUMP)

    @property
    def provurl(self):
        """
        The url where the prov document is saved
        """
        return url.join(self.tmpurl, self.PROV_DUMP)

    @property
    def extrainfourl(self):
        """
        The url where the extra job info document is saved
        """
        return url.join(self.tmpurl, self.INFO_DUMP)

    def get_deferred(self, output_id, cardinality_nr, sample_id=None):
        """
        Get a deferred pointing to a specific output value in the Job

        :param str output_id: the output to select from
        :param int cardinality_nr: the index of the cardinality
        :param str sample_id: the sample id to select (optional)
        :return: The deferred
        """
        parsed_url = urlparse.urlparse(self.logurl)

        query = {
            'outputname': output_id,
            'nr': cardinality_nr,
        }

        if sample_id is not None:
            query['sampleid'] = sample_id

        deffered_url = urlparse.urlunparse(
            (
                'val',
                parsed_url.netloc,
                parsed_url.path,
                parsed_url.params,
                urllib.urlencode(query),
                ''
            )
        )

        return Deferred(deffered_url)

    @property
    def commandfile(self):
        """
        The path of the command pickle
        """
        return fastr.vfs.url_to_path(self.commandurl)

    @property
    def logfile(self):
        """
        The path of the result pickle
        """
        return fastr.vfs.url_to_path(self.logurl)

    @property
    def stdoutfile(self):
        """
        The path where the stdout text is saved
        """
        return fastr.vfs.url_to_path(self.stdouturl)

    @property
    def stderrfile(self):
        """
        The path where the stderr text is saved
        """
        return fastr.vfs.url_to_path(self.stderrurl)

    @property
    def provfile(self):
        """
        The path where the prov document is saved
        """
        return fastr.vfs.url_to_path(self.provurl)

    @property
    def extrainfofile(self):
        """
        The path where the extra job info document is saved
        """
        return fastr.vfs.url_to_path(self.extrainfourl)

    @property
    def required_cores(self):
        """
        Number of required cores
        """
        return self._required_cores

    @required_cores.setter
    def required_cores(self, value):
        """
        Number of required cores
        """
        if value is None:
            self._required_cores = value
        else:
            if not isinstance(value, int):
                raise TypeError('Required number of cores should be an integer or None')

            if value < 1:
                raise ValueError('Required number of cores should be above zero ({} < 1)'.format(value))

            self._required_cores = value

    @property
    def required_memory(self):
        """
        Number of required memory
        """
        return self._required_memory

    @required_memory.setter
    def required_memory(self, value):
        """
        Number of required memory
        """
        if value is None:
            self._required_memory = value
        else:
            if isinstance(value, unicode):
                value = str(value)

            if not isinstance(value, str):
                raise TypeError('Required memory should be a str or None (found: {} [{}])'.format(value, type(value).__name__))

            if re.match(r'\d+[mMgG]', value) is None:
                raise ValueError('Required memory should be in the form \\d+[mMgG] (found {})'.format(value))

            self._required_memory = value

    @property
    def required_time(self):
        """
        Number of required runtime
        """
        return self._required_time

    @required_time.setter
    def required_time(self, value):
        """
        Number of required runtime
        """
        if value is None:
            self._required_time = value
        else:
            if isinstance(value, unicode):
                value = str(value)

            if not isinstance(value, str):
                raise TypeError('Required number of cores should be a str or None')

            if re.match(r'^(\d*:\d*:\d*|\d+)$', value) is None:
                raise ValueError('Required memory should be in the form HH:MM:SS or MM:SS (found {})'.format(value))

            self._required_time = value

    @property
    def tool(self):
        return fastr.toollist[self.tool_name, self.tool_version]

    def ensure_tmp_dir(self):
        # Determine absolute location of output dir and create directory
        output_dir = url.get_path_from_url(self.tmpurl)

        # Remove output directory if there is old stuff present
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        if not os.path.exists(output_dir):
            fastr.log.critical('Could not create output directory {}!'.format(output_dir))

    def create_payload(self):
        """
        Create the payload for this object based on all the input/output
        arguments

        :return: the payload
        :rtype: dict
        """
        tool = self.tool
        payload = {'inputs': {}, 'outputs': {}}

        # Fill the payload with the values to use (these should be translated to paths/strings/int etc
        # Translate all inputs to be in correct form
        for id_, value in self.input_arguments.items():
            argument = tool.inputs[id_]
            if isinstance(value, SampleItem):
                if len(value.data.mapping_part()) == 0:
                    value = value.data.sequence_part()
                elif len(value.data.sequence_part()) == 0:
                    value = value.data.mapping_part()
                else:
                    raise ValueError('Fastr does not (yet) accept mixed sequence/mapping input!')

            if not argument.hidden:
                if isinstance(value, tuple):
                    payload['inputs'][id_] = tuple(self.translate_argument(x) for x in value)
                else:  # Should be ordered dict
                    # FIXME: v is actually a tuple that needs fixing
                    payload['inputs'][id_] = OrderedDict((k, tuple(self.translate_argument(x) for x in v)) for k, v in value.items())

            else:
                if issubclass(fastr.typelist[argument.datatype], URLType):
                    payload['inputs'][id_] = tuple(self.translate_argument(x) for x in value)
                else:
                    payload['inputs'][id_] = value

            if len(payload['inputs'][id_]) == 0 and argument.default is not None:
                payload['inputs'][id_] = (argument.default,)

        # Create output arguments automatically
        for id_, spec in self.output_arguments.items():
            argument = tool.outputs[id_]

            if not argument.automatic:
                if isinstance(spec['cardinality'], int):
                    cardinality = spec['cardinality']
                else:
                    cardinality = self.calc_cardinality(spec['cardinality'], payload)
            else:
                cardinality = 1

            payload['outputs'][id_] = self.fill_output_argument(tool.outputs[id_],
                                                                cardinality,
                                                                spec['datatype'],
                                                                spec['requested'])

        return payload

    @staticmethod
    def calc_cardinality(description, payload):
        if description == 'unknown':
            return None

        if isinstance(description, int):
            return description

        if isinstance(description, str):
            description = Output.create_output_cardinality(description)

        if description[0] == 'int':
            return description[1]
        elif description[0] == 'as':
            if description[1] in payload['inputs']:
                return len(payload['inputs'][description[1]])
            if description[1] in payload['outputs']:
                return len(payload['outputs'][description[1]])
            else:
                raise exceptions.FastrValueError('Cannot determine cardinality from {} (payload {})'.format(description, payload))
        elif description[0] == 'val':
            if description[1] in payload['inputs'] and len(payload['inputs'][description[1]]) == 1:
                return int(str(payload['inputs'][description[1]][0]))
            if description[1] in payload['outputs'] and len(payload['outputs'][description[1]]) == 1:
                return int(str(payload['outputs'][description[1]][0]))
            else:
                raise exceptions.FastrValueError('Cannot determine cardinality from {} (payload {})'.format(description, payload))
        else:
            raise exceptions.FastrValueError('Cannot determine cardinality from {} (payload {})'.format(description, payload))

    def execute(self):
        """
        Execute this job

        :returns: The result of the execution
        :rtype: InterFaceResult
        """
        # Check if Fastr version is stored or we use a version that predates
        # this feature
        if not hasattr(self, 'fastr_version'):
            message = (
                'Job was created with an old version of Fastr that did not'
                'track the fastr_version yet, current Fastr version is {}'
            ).format(fastr.version.full_version)

            # For default branch (production) this is an error, for development
            # just a warning
            if fastr.version.hg_branch == 'default':
                fastr.log.critical(message)
                raise exceptions.FastrVersionMismatchError(message)
            else:
                fastr.log.warning(message)

        # Check if the fastr version is identical to the version that created
        # the job initially.
        if fastr.version.full_version != self.fastr_version:
            message = (
                'Job was created using Fastr version {}, but execution is'
                ' attempted on Fastr version {}'
            ).format(
                self.fastr_version,
                fastr.version.full_version
            )

            # For default branch (production) this is an error, for development
            # just a warning
            if fastr.version.hg_branch == 'default':
                fastr.log.critical(message)
                raise exceptions.FastrVersionMismatchError(message)
            else:
                fastr.log.warning(message)

        # Change the working directory to job temp dir
        old_curdir = os.path.abspath(os.curdir)
        job_dir = fastr.vfs.url_to_path(self.tmpurl)
        fastr.log.info('Set current directory to job output dir {}'.format(job_dir))
        os.chdir(job_dir)

        tool = fastr.toollist[self.tool_name, self.tool_version]

        # Hash the inputs
        self.hash_inputs()

        # Create the payload
        fastr.log.info('Start executing tool')
        start = datetime.datetime.utcnow()
        payload = self.create_payload()
        end = datetime.datetime.utcnow()
        fastr.log.info('Finished creating payload in {} seconds'.format((end - start).total_seconds()))

        # Execute the tool
        fastr.log.info('Start executing tool')
        start = datetime.datetime.utcnow()
        result = tool.execute(payload)
        end = datetime.datetime.utcnow()
        fastr.log.info('Finished executing tool in {} seconds'.format((end - start).total_seconds()))

        # Save the log data
        self.info_store['process'] = result.log_data

        # Check if there were errors in the Interface result
        self.errors.extend(result.errors)

        fastr.log.info('Start translating results tool')
        start = datetime.datetime.utcnow()
        self.output_data = self.translate_results(result.result_data)
        end = datetime.datetime.utcnow()
        fastr.log.info('Finished translating results in {} seconds'.format((end - start).total_seconds()))

        # Collect the provenance for the node
        self.provenance.collect_provenance(self)

        # Return the working directory to the old state
        fastr.log.info('Resetting current directory to {}'.format(old_curdir))
        os.chdir(old_curdir)

        if not self.validate_results(payload):
            raise exceptions.FastrValueError('Output values are not valid!')

        return result

    def translate_argument(self, value):
        """
        Translate an argument from a URL to an actual path.

        :param value: value to translate
        :param datatype: the datatype of the value
        :return: the translated value
        """
        return self.get_value(value=value)

    def get_output_datatype(self, output_id):
        """
        Get the datatype for a specific output

        :param str output_id: the id of the output to get the datatype for
        :return: the requested datatype
        :rtype: BaseDataType
        """
        output = self.tool.outputs[output_id]
        datatype = fastr.typelist[output.datatype]

        # If there are preferred types, match with that if possible
        if output_id in self.preferred_types and len(self.preferred_types[output_id]) > 0:
            new_datatype = fastr.typelist.match_types(datatype, preferred=self.preferred_types[output_id])
            if new_datatype is not None:
                fastr.log.info('Found new type (after using preferred): {} -> {}'.format(datatype.id, new_datatype.id))
                datatype = new_datatype
        return datatype

    def translate_results(self, result):
        """
        Translate the results of an interface (using paths etc) to the proper
        form using URI's instead.

        :param dict result: the result data of an interface
        :return: the translated result
        :rtype: dict
        """
        for key, value in result.items():
            datatype = self.get_output_datatype(key)

            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    new_subvalue = []
                    for item in subvalue:
                        if not isinstance(item, DataType):
                            item = datatype(str(item))
                        if isinstance(item, URLType):
                            item.value = fastr.vfs.path_to_url(item.value)
                        new_subvalue.append(item)
                    value[subkey] = tuple(new_subvalue)
            else:
                new_value = []
                for item in value:
                    if not isinstance(item, DataType):
                        item = datatype(str(item))
                    if isinstance(item, URLType):
                        item.value = fastr.vfs.path_to_url(item.value)
                    new_value.append(item)
                result[key] = new_value

        return result

    def fill_output_argument(self, output_spec, cardinality, desired_type, requested):
        """
        This is an abstract class method. The method should take the argument_dict
        generated from calling self.get_argument_dict() and turn it into a list
        of commandline arguments that represent this Input/Output.

        :param int cardinality: the cardinality for this output (can be non for automatic outputs)
        :param DataType desired_type: the desired datatype for this output
        :param bool requested: flag to indicate that the output is requested by Fastr
        :return: the values for this output
        :rtype: list
        """
        values = []

        if not output_spec.automatic:
            datatype = fastr.typelist[desired_type]

            for cardinality_nr in range(cardinality):
                if datatype.extension is not None:
                    output_url = '{}/{}_{}.{}'.format(self.tmpurl, output_spec.id, cardinality_nr, datatype.extension)
                else:
                    output_url = '{}/{}_{}'.format(self.tmpurl, output_spec.id, cardinality_nr)

                # Wrap the output url in the correct DataType
                fastr.log.debug('Wrapping {} in a {}'.format(output_url, datatype))
                output_value = datatype(output_url)
                fastr.log.debug('Got {}'.format(output_value))

                # Translate to a path and use
                values.append(self.translate_argument(output_value))
        else:
            datatype = fastr.typelist['Boolean']
            values.append(datatype(requested))

        return tuple(values)

    @classmethod
    def get_value(cls, value):
        """
        Get a value

        :param value: the url of the value
        :param datatype: datatype of the value
        :return: the retrieved value
        """
        if isinstance(value, Deferred):
            value = value.target

        # If the value already has valid datatype, use that and don't guess from scratch
        if isinstance(value, URLType):
            datatype = type(value)
            if url.isurl(str(value)):
                value = fastr.vfs.url_to_path(str(value))
            else:
                if not os.path.exists(str(value)):
                    raise exceptions.FastrValueError('Found a non-url path ({}) that does not exist!'.format(value))
            return datatype(value)
        elif isinstance(value, DataType):
            return value
        else:
            raise exceptions.FastrTypeError('Arguments should be a DataType, found {}'.format(type(value).__name__))

    def hash_inputs(self):
        """
        Create hashes for all input values and store them in the info store
        """
        for key, value in self.input_arguments.items():
            if value.data.is_sequence:
                self.input_hash[key] = [x.checksum() for x in value.data.sequence_part()]
            else:
                self.input_hash[key] = {}
                for sample_id, input_val in value.data.mapping_part().items():
                    self.input_hash[key][sample_id] = [x.checksum() for x in input_val]

    def hash_results(self):
        """
        Create hashes of all output values and store them in the info store
        """
        for output in self.output_arguments.values():
            if not output['requested'] and output['id'] not in self.output_data:
                continue

            id_ = output['id']
            output_value = self.output_data[id_]

            if isinstance(output_value, list):
                self.output_hash[id_] = [x.checksum() if x.valid else None for x in output_value]
            elif isinstance(output_value, dict):
                self.output_hash[id_] = {}
                for sample_id, output_val in output_value.items():
                    self.output_hash[id_][sample_id] = [x.checksum() if x.valid else None for x in output_val]

    def validate_results(self, payload):
        """
        Validate the results of the Job

        :return: flag indicating the results are complete and valid
        """

        valid = True
        for output in self.tool.outputs.values():
            id_ = output.id

            if id_ not in self.output_data:
                if id_ in self.output_arguments and self.output_arguments[id_]['requested']:
                    message = 'Could not find result for output {}'.format(id_)
                    self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                    fastr.log.warning(message)
                    valid = False
                else:
                    fastr.log.warning('Could not find non-required output {}'.format(id_))
                continue

            output_value = self.output_data[id_]

            if isinstance(output_value, (list, tuple)):
                if not self._validate_result(output, output_value, payload):
                    message = 'The output "{}" is invalid!'.format(id_)
                    self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                    fastr.log.warning(message)
                    valid = False
            elif isinstance(output_value, dict):
                for sample_id, output_val in output_value.items():
                    if not self._validate_result(output, output_val, payload):
                        message = 'The output "{}" for sample "{}" is invalid!'.format(id_, sample_id)
                        self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                        fastr.log.warning(message)
                        valid = False

            else:
                raise exceptions.FastrTypeError('Output data is not of correct type (expected a list/dict)')

        return valid

    def write(self):
        self.provenance.serialize(self.provfile, 'json')
        save_json(self.extrainfofile, self.info_store)
        save_gpickle(self.logfile, self)

    def _validate_result(self, output, output_value, payload):
        """
        Validate the result for a specific otuput/sample
        :param output: the output for which to check
        :param output_value: the value for the output
        :return: flag indicating if the result is value
        """
        valid = True
        cardinality = self.calc_cardinality(output.cardinality, payload)

        if cardinality is not None and cardinality != len(output_value):
            message = 'Cardinality mismatch for {} (found {}, expected {})'.format(output.id,
                                                                                   len(output_value),
                                                                                   cardinality)
            self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
            fastr.log.warning(message)
            valid = False

        for value in output_value:
            if not value.valid:
                message = 'Output value [{}] "{}" not valid for datatype "{}"'.format(type(value).__name__,
                                                                                      value,
                                                                                      output.datatype)
                self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                fastr.log.warning(message)
                valid = False

        fastr.log.info('Data for output {} is {}'.format(output.id, 'valid' if valid else 'invalid'))

        return valid


class SinkJob(Job):
    """
    Special SinkJob for the Sink
    """
    def __init__(self, node, sample_id, sample_index, input_arguments,
                 output_arguments, hold_jobs=None, substitutions=None,
                 status_callback=None, preferred_types=None):

        self.cardinality = substitutions['cardinality']  # This is required!
        super(SinkJob, self).__init__(node=node,
                                      sample_id=sample_id,
                                      sample_index=sample_index,
                                      input_arguments=input_arguments,
                                      output_arguments=output_arguments,
                                      hold_jobs=hold_jobs,
                                      status_callback=status_callback,
                                      preferred_types=preferred_types)

        self._substitutions = substitutions if substitutions else {}

    def __repr__(self):
        """
        String representation for the SinkJob
        """
        return '<SinkJob\n  id={job.id}\n  tmpdir={job.tmpurl}/>'.format(job=self)

    @property
    def id(self):
        """
        The id of this job
        """
        return '{}___{}___{}___{}'.format(
            self.network_id,
            self.node_id,
            self.sample_id,
            self.cardinality
        )

    @property
    def tmpurl(self):
        """
        The URL of the tmpdir to use
        """
        return '{}/{}/{}_{}'.format(self.network_tmpurl, self.node_id,
                                    self.sample_id, self.cardinality)

    def get_result(self):
        """
        Get the result of the job if it is available. Load the output file if
        found and check if the job matches the current object. If so, load and
        return the result.

        :returns: Job after execution
        """
        return None

    def validate_results(self, payload):
        """
        Validate the results of the SinkJob

        :return: flag indicating the results are complete and valid
        """
        if self.info_store['process']['stderr'] != '':
            message = 'SinkJob should have an empty stderr, found error messages!\n{}'.format(self.info_store['process']['stderr'])
            fastr.log.warning(message)
            self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
            return False
        else:
            return True

    def create_payload(self):
        """
        Create the payload for this object based on all the input/output
        arguments

        :return: the payload
        :rtype: dict
        """
        payload = super(SinkJob, self).create_payload()
        fastr.log.info('Temp payload: {}'.format(payload))
        fastr.log.info('Input data: {}'.format(payload['inputs']['input']))
        fastr.log.info('Input arguments: {}'.format(self.input_arguments['input'].data))

        payload['inputs']['output'] = tuple(self.substitute(in_, datatype=type(out)) for in_, out in zip(payload['inputs']['output'],
                                                                                                         payload['inputs']['input']))

        # Add the provenance information into the mix
        extra_inputs = []
        extra_outputs = []
        extra_datatypes = []

        if len(self.input_arguments['input'].data.mapping_part()) > 0:
            raise exceptions.FastrValueError('SinkJob should not have input with mapping_parts!')

        inputs = self.input_arguments['input'].data.sequence_part()
        outputs = payload['inputs']['output']

        for input_, output in zip(inputs, outputs):
            fastr.log.info('Checking [{}] {!s} ({!r})'.format(type(input_), input_, input_))
            if isinstance(input_, Deferred):
                parsed_url = urlparse.urlparse(input_.raw_value)

                if parsed_url.scheme != 'val':
                    raise exceptions.FastrValueError('Cannot lookup value {}, wrong url scheme'.format(value.raw_value))

                # Find location of the provn file
                datafile = os.path.join(fastr.config.mounts[parsed_url.netloc],
                                        os.path.normpath(parsed_url.path[1:]))
                datafile = datafile.replace(self.RESULT_DUMP, self.PROV_DUMP)

                extra_inputs.append(fastr.datatypes.JsonFile(datafile))
                extra_outputs.append(output + '.prov.json')
                extra_datatypes.append(fastr.datatypes.String('JsonFile'))

        fastr.log.info('Extra inputs: {}  Extra outputs: {}'.format(extra_inputs,
                                                                    extra_outputs))

        payload['inputs']['input'] = payload['inputs']['input'] + tuple(extra_inputs)
        payload['inputs']['output'] = payload['inputs']['output'] + tuple(extra_outputs)
        payload['inputs']['datatype'] = payload['inputs']['datatype'] + tuple(extra_datatypes)

        return payload

    def substitute(self, value, datatype=None):
        """
        Substitute the special fields that can be used in a SinkJob.

        :param str value: the value to substitute fields in
        :param BaseDataType datatype: the datatype for the value
        :return: string with substitutions performed
        :rtype: str
        """
        if datatype is None:
            datatype = type(value)

        subs = dict(self._substitutions)
        subs['ext'] = '.{}'.format(datatype.extension) if datatype.extension is not None else ''
        subs['extension'] = '{}'.format(datatype.extension) if datatype.extension is not None else ''
        return str(value).format(**subs)

    def hash_inputs(self):
        """
        Create hashes for all input values and store them in the info store
        """
        input_data = self.input_arguments['input'].data
        if input_data.is_sequence:
            self.input_hash['input'] = [x.checksum() for x in input_data.sequence_part()]
        else:
            self.input_hash['input'] = {}
            for sample_id, input_val in input_data.mapping_part().items():
                self.input_hash['input'][sample_id] = [x.checksum() for x in input_val]


class SourceJob(Job):
    """
    Special SourceJob for the Source
    """
    def __repr__(self):
        """
        String representation for the SourceJob
        """
        return '<SourceJob\n  id={job.id}\n  tmpdir={job.tmpurl}/>'.format(job=self)

    def validate_results(self, payload):
        """
        Validate the results of the Job

        :return: flag indicating the results are complete and valid
        """
        if self.info_store['process']['stderr'] != '':
            message = 'SourceJob should have an empty stderr, found error messages!'
            fastr.log.warning(message)
            self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
            return False
        else:
            return True

    def get_output_datatype(self, output_id):
        """
        Get the datatype for a specific output

        :param str output_id: the id of the output to get the datatype for
        :return: the requested datatype
        :rtype: BaseDataType
        """
        return fastr.typelist[self._datatype]

    def hash_inputs(self):
        """
        Create hashes for all input values and store them in the info store
        """
        inputs = [fastr.datatypes.String(x) for x in self.input_arguments['input']]
        self.input_hash['input'] = [x.checksum() for x in inputs]


class InlineJob(Job):
    def __init__(self, *args, **kwargs):
        super(InlineJob, self).__init__(*args, **kwargs)
        self.provenance.init_provenance(self)
        self.provenance.collect_provenance(self, advanced_flow=True)

    def get_result(self):
        if not os.path.exists(self.logfile):
            return None

        try:
            result = load_gpickle(self.logfile)
        except (IOError, EOFError):
            # Errors loading pickle or gzip stream
            return None

        if not isinstance(result, Job):
            fastr.log.debug('Result is not valid Job! (found {})'.format(type(result).__name__))
            return None

        return result



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The module containing the classes describing the targets.
"""

from abc import ABCMeta, abstractmethod
from collections import deque, namedtuple, Sequence
import datetime
import os
import platform
import psutil
import subprocess
import time
import threading
import shellescape

import isodate
import requests

import fastr
from fastr import exceptions
from fastr.data import url

# Check if docker is available
try:
    import docker
    DOCKER_AVAILABLE = True
except ImportError:
    docker = None
    DOCKER_AVAILABLE = False

# Check if environment modules are available
try:
    from fastr.execution.environmentmodules import EnvironmentModules
    ENVIRONMENT_MODULES = EnvironmentModules(fastr.config.protected_modules)
    ENVIRONMENT_MODULES_LOADED = True
except exceptions.FastrValueError:
    ENVIRONMENT_MODULES = None
    ENVIRONMENT_MODULES_LOADED = False

try:
    from nipype.interfaces import base as nipypebase
    NIPYPE_AVAILABLE = True
except:
    NIPYPE_AVAILABLE = False

# Monitor interval for profiling
_MONITOR_INTERVAL = 1.0

SystemUsageInfo = namedtuple('SystemUsageInfo', ['timestamp',
                                                 'cpu_percent',
                                                 'vmem',
                                                 'rmem',
                                                 'read_bytes',
                                                 'write_bytes'])


class ProcessUsageCollection(Sequence):
    # It has to be defined in module for pickling purposes
    usage_type = SystemUsageInfo

    def __init__(self):
        self.seconds_info = deque()
        self.minutes_info = []

    def __len__(self):
        return len(self.seconds_info) + len(self.minutes_info)

    def __getitem__(self, item):
        # First look in minutes, then in seconds
        if item < len(self.minutes_info):
            return self.minutes_info[item]._asdict()
        else:
            return self.seconds_info[item - len(self.minutes_info)]._asdict()

    def append(self, value):
        if not isinstance(value, self.usage_type):
            raise ValueError('Cannot add a non {}.usage_type'.format(type(self).__name__))

        self.seconds_info.append(value)

        if len(self.seconds_info) >= 120:
            self.aggregate(60)

    def aggregate(self, number_of_points):
        oldest_data = [self.seconds_info.popleft() for _ in range(number_of_points)]

        timestamp = oldest_data[-1].timestamp
        cpu_percent = sum(x.cpu_percent for x in oldest_data) / len(oldest_data)
        vmem = max(x.vmem for x in oldest_data)
        rmem = max(x.rmem for x in oldest_data)
        read_bytes = oldest_data[-1].read_bytes
        write_bytes = oldest_data[-1].write_bytes

        self.minutes_info.append(self.usage_type(timestamp=timestamp,
                                                 cpu_percent=cpu_percent,
                                                 vmem=vmem,
                                                 rmem=rmem,
                                                 read_bytes=read_bytes,
                                                 write_bytes=write_bytes))


class Target(object):
    """
    The abstract base class for all targets. Execution with a target should
    follow the following pattern:

    >>> with Target() as target:
    ...     target.run_commmand(['sleep', '10'])
    ...     target.run_commmand(['sleep', '10'])
    ...     target.run_commmand(['sleep', '10'])

    The Target context operator will set the correct paths/initialization.
    Within the context command can be ran and when leaving the context the
    target reverts the state before.
    """
    __metaclass__ = ABCMeta

    if NIPYPE_AVAILABLE:
        _NIPYPE_RUN_COMMAND = nipypebase.run_command

        def nipype_run(self, runtime, output, timeout=None, *args, **kwargs):
            """
            A command that has the same signature as the nipype.interfaces.base.run_command

            This adapts the call to the self.run_command that fastr uses for just dispatching
            without environment setting.
            """
            # It is safe to ignore the environment (as it is just a copy)
            # It is safe to ignore cwd (as it is just a copy)
            # See nipype.interfaces.base:974
            fastr.log.info('Running nipype interface using fastr-patched run_command')
            if len(args) > 0:
                fastr.log.warning('Found uncaught args: {}'.format(args))
            if len(kwargs) > 0:
                fastr.log.warning('Found uncaught kwargs: {}'.format(kwargs))

            result = self.run_command(runtime.cmdline.split())

            # Copy the resulting data into the runtime Bunch
            runtime.stdout = result['stdout']
            runtime.stderr = result['stderr']
            runtime.merged = ''
            runtime.returncode = result['returncode']

            return runtime
    else:
        _NIPYPE_RUN_COMMAND = None

    def __enter__(self):
        """
        Set the environment in such a way that the target will be on the path.
        """
        if NIPYPE_AVAILABLE:
            # Make sure nipype runs via the Target and not just spawn own subprocesses
            nipypebase.run_command = self.nipype_run
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleanup the environment where needed
        """
        if NIPYPE_AVAILABLE:
            # Reset nipype to work with own run_command
            nipypebase.run_command = self._NIPYPE_RUN_COMMAND

    @abstractmethod
    def run_command(self, command):
        pass


class LocalBinaryTarget(Target):
    """
    A tool target that is a local binary on the system. Can be found using
    environmentmodules or vfs-path on the executing machine
    """

    DYNAMIC_LIBRARY_PATH_DICT = {
        'windows': 'PATH',  # Not Tested
        'linux': 'LD_LIBRARY_PATH',  # Tested
        'darwin': 'DYLD_LIBRARY_PATH',  # Tested
    }

    _platform = platform.system().lower()
    if _platform not in DYNAMIC_LIBRARY_PATH_DICT:
        fastr.log.warning('"Dynamic library path not supported on platform: {}"'.format(_platform))

    def __init__(self, binary, paths=None, environment_variables=None,
                 initscripts=None, modules=None, interpreter=None, **kwargs):
        """
        Define a new local binary target. Must be defined either using paths and optionally environment_variables
        and initscripts, or enviroment modules.
        """
        self.binary = binary
        if modules is None:
            if 'module' in kwargs and kwargs['module'] is not None:
                fastr.log.warning('Using deprecated module in target (modules is new way to do it)')
                self._modules = (kwargs['module'],)
            else:
                self._modules = None
        elif isinstance(modules, str):
            self._modules = (modules.strip(),)
        else:
            self._modules = tuple(x.strip() for x in modules)

        if isinstance(paths, str):
            self._paths = [{'type': 'bin', 'value': paths}]
        elif paths is None and 'location' in kwargs and kwargs['location'] is not None:
            fastr.log.warning('Using deprecated location in target (paths is the new way to do it)')
            self._paths = [{'type': 'bin', 'value': kwargs['location']}]
        else:
            self._paths = paths

        if paths is not None:
            for path_entry in self._paths:
                if not url.isurl(path_entry['value']):
                    path_entry['value'] = os.path.abspath(path_entry['value'])

        if environment_variables is None:
            environment_variables = {}
        self._envvar = environment_variables

        if initscripts is None:
            initscripts = []
        self._init_scripts = initscripts

        self.interpreter = interpreter

        self._roll_back = None

    def __enter__(self):
        """
        Set the environment in such a way that the target will be on the path.
        """
        super(LocalBinaryTarget, self).__enter__()

        # Create dictionary of possible platforms, to set dynamic labrary path
        # Add check to see if _platform is present in dictionary
        if self._platform in self.DYNAMIC_LIBRARY_PATH_DICT:
            dynamic_library_path = self.DYNAMIC_LIBRARY_PATH_DICT[self._platform]
        else:
            dynamic_library_path = None

        if ENVIRONMENT_MODULES_LOADED and self._modules is not None and len(self._modules) > 0:
            # Clear the enviroment modules and load all required modules
            ENVIRONMENT_MODULES.clear()
            for module_ in self._modules:
                if not ENVIRONMENT_MODULES.isloaded(module_):
                    ENVIRONMENT_MODULES.load(module_)
                    fastr.log.info('loaded module: {}'.format(module_))
            fastr.log.info('LoadedModules: {}'.format(ENVIRONMENT_MODULES.loaded_modules))
        elif self._paths is not None:
            # Prepend PATH and LD_LIBRARY_PATH as required
            self._roll_back = {'PATH': os.environ.get('PATH', None)}

            # Prepend extra paths to PATH
            bin_path = os.environ.get('PATH', None)
            bin_path = [bin_path] if bin_path else []
            extra_path = [x['value'] for x in self._paths if x['type'] == 'bin']
            extra_path = [fastr.vfs.url_to_path(x) if url.isurl(x) else x for x in extra_path]
            fastr.log.info('Adding extra PATH: {}'.format(extra_path))
            os.environ['PATH'] = os.pathsep.join(extra_path + bin_path)

            # Prepend extra paths to LB_LIBRARY_PATH
            extra_ld_library_path = [x['value'] for x in self._paths if x['type'] == 'lib']
            if len(extra_ld_library_path) > 0:
                if dynamic_library_path is None:
                    message = 'Cannot set dynamic library path on platform: {}'.format(self._platform)
                    fastr.log.critical(message)
                    raise exceptions.FastrNotImplementedError(message)

                self._roll_back[dynamic_library_path] = os.environ.get(dynamic_library_path, None)

                lib_path = os.environ.get(dynamic_library_path, None)
                lib_path = [lib_path] if lib_path else []
                extra_ld_library_path = [fastr.vfs.url_to_path(x) if url.isurl(x) else x for x in extra_ld_library_path]

                fastr.log.info('Adding extra LIB: {}'.format(extra_path))
                os.environ[dynamic_library_path] = os.pathsep.join(extra_ld_library_path + lib_path)

            # Set other environment variables as indicated
            for var, value in self._envvar.items():
                if var in ['PATH', dynamic_library_path]:
                    continue

                self._roll_back[var] = os.environ.get(var, None)
                os.environ = value

            # Run init script(s) if required
            for script in self._init_scripts:
                if isinstance(script, str):
                    script = [script]

                subprocess.call(script)
        else:
            raise exceptions.FastrNotImplementedError(
                'Binary targets must have either paths or modules set! (binary {})'.format(self.binary)
            )

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleanup the environment
        """
        if ENVIRONMENT_MODULES_LOADED and self._modules is not None and len(self._modules) > 0:
            ENVIRONMENT_MODULES.clear()
        elif self._roll_back is not None:
            for var, value in self._roll_back.items():
                if value is not None:
                    os.environ[var] = value
                else:
                    del os.environ[var]

            self._roll_back = None

    def call_subprocess(self, command):
        """
        Call a subprocess with logging/timing/profiling

        :param list command: the command to execute
        :return: execution info
        :rtype: dict
        """
        sysuse = ProcessUsageCollection()
        start_time = time.time()
        fastr.log.info('Calling command arguments: {}'.format(command))
        printable_command = []
        for item in command:
            printable_command.append(shellescape.quote(item))
        fastr.log.info('Calling command: "{}"'.format(' '.join(printable_command)))
        try:
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        except OSError as exception:
            if exception.errno == 2:
                raise exceptions.FastrExecutableNotFoundError(command[0])
            elif exception.errno == 13:
                # Permission denied
                raise exceptions.FastrNotExecutableError('Cannot execute {}, permission denied!'.format(command[0]))
            else:
                raise exception
        monitor_thread = threading.Thread(target=self.monitor_process, name='SubprocessMonitor', args=(process, sysuse))
        monitor_thread.daemon = True  # Make sure this Thread does not block exiting the script
        monitor_thread.start()
        stdout, stderr = process.communicate()
        return_code = process.poll()
        end_time = time.time()

        if monitor_thread.is_alive():
            monitor_thread.join(2 * _MONITOR_INTERVAL)
            if monitor_thread.is_alive():
                fastr.log.warning('Ignoring unresponsive monitor thread!')

        return {'returncode': return_code,
                'stdout': stdout,
                'stderr': stderr,
                'command': command,
                'resource_usage': list(sysuse),
                'time_elapsed': end_time - start_time}

    def monitor_process(self, process, resources):
        """
        Monitor a process and profile the cpu, memory and io use. Register the
        resource use every _MONITOR_INTERVAL seconds.

        :param subproces.Popen process: process to monitor
        :param resources: list to append measurements to
        """
        psproc = psutil.Process(process.pid)

        # Loop initialization
        psproc.cpu_percent()  # Get rid of meaningless 0.0 at start
        last_timestamp = datetime.datetime.utcnow()

        while process.poll() is None:
            try:
                # The sleep duration is adapted to loop duration so aggregation will not cause
                # extended intervals
                sleep_duration = _MONITOR_INTERVAL - (datetime.datetime.utcnow() - last_timestamp).total_seconds()
                sleep_duration = 0.0 if sleep_duration < 0.0 else sleep_duration
                time.sleep(sleep_duration)

                # Get process usage information
                memory_info = psproc.memory_info()

                if self._platform == 'darwin':
                    io_read = None
                    io_write = None
                else:
                    io_info = psproc.io_counters()
                    io_read = io_info.read_bytes
                    io_write = io_info.write_bytes

                last_timestamp = datetime.datetime.utcnow()
                usage = resources.usage_type(timestamp=last_timestamp.isoformat(),
                                             cpu_percent=psproc.cpu_percent(),
                                             vmem=memory_info.vms,
                                             rmem=memory_info.rss,
                                             read_bytes=io_read,
                                             write_bytes=io_write)

                resources.append(usage)

            except psutil.Error:
                # If the error occured because during the interval of meassuring the CPU use
                # the process stopped, we do not mind
                if process.poll() is None:
                    raise

    def run_command(self, command):
        if self.interpreter is not None:
            paths = [x['value'] for x in self._paths if x['type'] == 'bin']
            fastr.log.info('Options: {}'.format(paths))
            containing_path = next(x for x in paths if os.path.exists(os.path.join(x, command[0])))
            command = [self.interpreter, os.path.join(containing_path, command[0])] + command[1:]

        fastr.log.debug('COMMAND: "{}" ({})'.format(command, type(command).__name__))
        return self.call_subprocess(command)


class DockerTarget(Target):
    """
    A tool target that is located in a Docker images. Can be run using
    docker-py.
    """
    def __init__(self, binary, docker_image):
        """
        Define a new docker target.

        :param str docker_image: Docker image to use
        """
        if not DOCKER_AVAILABLE:
            raise exceptions.FastrOptionalModuleNotAvailableError('Target cannot be used, module "docker" unavailable')

        self.binary = binary
        self._docker_image = docker_image

        #: Docker api to use for docker target
        self.docker_api = 'unix://var/run/docker.sock'

        self._docker_client = docker.DockerClient(base_url=self.docker_api,
                                                  version='auto')
        self._container = None

    def __enter__(self):
        super(DockerTarget, self).__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        self._container = None

    @property
    def container(self):
        return self._container

    def run_command(self, command):
        # Bind all fastr mounts
        mounts = fastr.config.mounts.values()
        fastr.log.info('DOCKER MOUNTS: {}'.format(mounts))
        binds = {x: {'bind': x, 'mode': 'ro'} for x in mounts if os.path.exists(x)}
        binds[fastr.config.mounts['tmp']]['mode'] = 'rw'
        fastr.log.info('DOCKER BINDS: {}'.format(binds))

        fastr.log.info('DOCKER CREATE CONTAINER')
        container = self._docker_client.containers.run(
            image=self._docker_image,
            command=command,
            detach=True,
            volumes=binds,
            network_disabled=True,
            user='{}:{}'.format(os.getuid(), os.getgid()),
            working_dir=os.path.abspath(os.curdir),
            environment=os.environ,
        )

        # Start monitoring
        sysuse = ProcessUsageCollection()
        monitor_thread = threading.Thread(target=self.monitor_docker,
                                          name='DockerMonitor',
                                          args=(container, sysuse))
        monitor_thread.daemon = True  # Make sure this Thread does not block exiting the script
        monitor_thread.start()
        start_time = time.time()

        # Run docker container
        fastr.log.info('DOCKER WAIT')
        return_code = container.wait()
        fastr.log.info('DOCKER WAIT DONE')
        end_time = time.time()
        stdout = container.logs(stdout=True, stderr=False, stream=False, timestamps=True)
        stderr = container.logs(stdout=False, stderr=True, stream=False, timestamps=True)

        if monitor_thread.is_alive():
            monitor_thread.join(2 * _MONITOR_INTERVAL)
            if monitor_thread.is_alive():
                fastr.log.warning('Ignoring unresponsive monitor thread!')

        # Clean the contianer
        container.remove()

        return {'returncode': return_code,
                'stdout': stdout,
                'stderr': stderr,
                'command': command,
                'resource_usage': list(sysuse),
                'time_elapsed': end_time - start_time}

    def monitor_docker(self, container, resources):
        """
        Monitor a process and profile the cpu, memory and io use. Register the
        resource use every _MONITOR_INTERVAL seconds.

        :param subproces.Popen process: process to monitor
        :param resources: list to append measurements to
        """
        try:
            for stat in container.stats(stream=True,
                                        decode=True):

                # Get cpu, memory and io statistics
                timestamp = isodate.isodatetime.parse_datetime(stat['read'])
                usage = resources.usage_type(timestamp=timestamp.isoformat(),
                                             cpu_percent=stat['cpu_stats']['cpu_usage']['total_usage'],
                                             vmem=-1.0,
                                             rmem=stat['memory_stats']['rss'],
                                             read_bytes=0.0,
                                             write_bytes=0.0)

                resources.append(usage)
        except requests.exceptions.ReadTimeout:
            fastr.log.info('Docker Monitor timed out')




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines the Fastr Config class for managing the configuration of
Fastr. The config object is stored directly in the fastr top-level module.
"""

import __builtin__
import collections
import inspect
import json
import logging
import logging.config
import multiprocessing
import os
import tempfile


logging.captureWarnings(True)


class EmptyDefault(object):
    def __init__(self, data=None):
        self._list_data = [NotImplemented]
        self._dict_data = collections.defaultdict(EmptyDefault)

        if isinstance(data, collections.Mapping):
            self._dict_data.update(data)
        elif isinstance(data, collections.Sequence):
            self._list_data = list(data)

    # Any object be added in place (just replaces current value)
    def __iadd__(self, right):
        self._list_data += right
        return self

    def __add__(self, right):
        return EmptyDefault(self._list_data + right)

    def __radd__(self, other):
        return EmptyDefault(other + self._list_data)

    def append(self, value):
        self._list_data.append(value)

    def prepend(self, value):
        self._list_data = [value] + self._list_data

    def extend(self, other):
        self._list_data.extend(other)

    # Make this act like a dict updated by another dict
    def update(self, other):
        self._dict_data.update(other)

    def __getitem__(self, item):
        return self._dict_data[item]

    def __setitem__(self, key, value):
        self._dict_data[key] = value

    def __delitem__(self, key):
        del self._dict_data[key]

    # Access to data
    def aslist(self):
        return list(self._list_data)

    def asdict(self):
        return dict(self._dict_data)


def _find_log_type():
    """
    Figure out the logtype to use for this fastr session

    :return: log type to us
    :rtype: str
    """
    # Use multiprocessing to check if we are actually in the MainProcess
    # Subprocesses should not log in a standard way
    current_process = multiprocessing.current_process()

    # Hack setting non-standard log methods/destinations before fastr is imported
    _stack = inspect.stack()
    for frame in _stack[1:]:
        if 'FASTR_LOG_TYPE' in frame[0].f_globals:
            fastr_log_type = frame[0].f_globals['FASTR_LOG_TYPE']

            # We only want to last definition of fastr_log_type
            break
    else:
        if current_process.name != 'MainProcess':
            #fastr_log_type = 'childprocess'
            fastr_log_type = 'default'
        else:
            fastr_log_type = 'default'

    return fastr_log_type


USER_DIR = os.path.expanduser(os.path.join('~', '.fastr'))
SYSTEM_DIR = os.path.abspath(os.path.normpath(os.path.dirname(__file__)))
RESOURCE_DIR = os.path.join(SYSTEM_DIR, 'resources')


class Config(object):
    """
    Class contain the fastr configuration
    """
    DEFAULT_FIELDS = {
        "debug": (
            bool,
            False,
            "Flag to enable/disable debugging"
        ),
        "logtype": (
            str,
            _find_log_type(),
            "Type of logging to use"
        ),
        "systemdir": (
            str,
            SYSTEM_DIR,
            "Fastr installation directory"
        ),
        "userdir": (
            str,
            USER_DIR,
            "Fastr user configuration directory",
            "~/.fastr"
        ),
        "logdir": (
            str,
            os.path.join(USER_DIR, 'logs'),
            "Directory where the fastr logs will be placed",
            "$userdir/logs"
        ),
        "resourcesdir": (
            str,
            RESOURCE_DIR,
            "Directory containing the fastr system resources",
            "$systemdir/resources"
        ),
        "examplesdir": (
            str,
            os.path.join(SYSTEM_DIR, 'examples'),
            "Directory containing the fastr examples",
            "$systemdir/examples"
        ),
        "schemadir": (
            str,
            os.path.join(SYSTEM_DIR, 'resources', 'schemas'),
            "Directory containing the fastr data schemas",
            "$systemdir/schemas"
        ),
        "executionscript": (
            str,
            os.path.join(SYSTEM_DIR, 'execution', 'executionscript.py'),
            "Execution script location",
            "$systemdir/execution/executionscript.py"
        ),
        "types_path": (
            list,
            [x for x in [os.path.join(USER_DIR, 'datatypes'), os.path.join(RESOURCE_DIR, 'datatypes')] if os.path.exists(x)],
            "Directories to scan for datatypes",
            ['$userdir/datatypes', '$resourcedir/datatypes']
        ),
        "tools_path": (
            list,
            [x for x in [os.path.join(USER_DIR, 'tools'), os.path.join(RESOURCE_DIR, 'tools')] if os.path.exists(x)],
            "Directories to scan for tools",
            ['$userdir/tools', '$resourcedir/tools']
        ),
        "networks_path": (
            list,
            [x for x in [os.path.join(USER_DIR, 'networks'), os.path.join(RESOURCE_DIR, 'networks')] if os.path.exists(x)],
            "Directories to scan for networks",
            ['$userdir/networks', '$resourcedir/networks']
        ),
        "plugins_path": (
            list,
            [x for x in [os.path.join(USER_DIR, 'plugins'), os.path.join(RESOURCE_DIR, 'plugins')] if os.path.exists(x)],
            "Directories to scan for plugins",
            ['$userdir/plugins', '$resourcedir/plugins']
        ),
        "mounts": (
            dict,
            {
                'tmp': tempfile.gettempdir(),
                'example_data': os.path.join(SYSTEM_DIR, 'examples', 'data'),
                'home': os.path.expanduser('~/'),
            },
            "A dictionary containing all mount points in the VFS system",
            {
                'tmp': '$TMPDIR',
                 'example_data': '$systemdir/examples/data',
                 'home': '~/'
            }
        ),
        "preferred_types": (
            list,
            [],
            "A list indicating the order of the preferred types to use. First item is most preferred."
        ),
        "protected_modules": (
            list,
            [],
            "A list of modules in the environmnet modules that are protected against unloading"
        ),
        "execution_plugin": (
            str,
            'ProcessPoolExecution',
            "The default execution plugin to use"
        ),
        "web_hostname": (
            str,
            'localhost',
            "The hostname to expose the web app for"
        ),
        "web_port": (
            str,
            '5000',
            "The port to expose the web app on"
        ),
        "web_secret_key": (
            str,
            'VERYSECRETKEY!',
            "The secret key to use for the flask web app"
        ),
        "warn_develop": (
            bool,
            True,
            "Warning users on import if this is not a production version of fastr"
        ),
        "source_job_limit": (
            int,
            4,
            "The number of source jobs allowed to run concurrently"
        ),
        "pim_host": (
            str,
            '',
            "Host of the PIM server to report to"
        )
    }

    # pylint: disable=too-many-instance-attributes
    # The config has many attributes, because its function is to hold
    # this data

    def __init__(self, *configfiles):
        #: Trace of the config files read by this object
        self.read_config_files = []
        self.log = None

        #: Raw config of current values
        self._fields = collections.OrderedDict(sorted(self.DEFAULT_FIELDS.items()))
        self._create_field_properties(self._fields)
        self._current_config = collections.defaultdict(EmptyDefault)
        self._current_config.update(vars(__builtin__))

        # Read default config files if found
        if os.environ.get('FASTRHOME'):
            if os.path.exists(os.path.join(os.environ['FASTRHOME'], 'config.py')):
                self.read_config(os.path.join(os.environ['FASTRHOME'], 'config.py'))

        if os.path.exists(os.path.join(self.userdir, 'config.py')):
            self.read_config(os.path.join(self.userdir, 'config.py'))

        # Read config files as parameters
        for filename in configfiles:
            if os.path.exists(filename):
                self.read_config(filename)
            else:
                self.log.error('Config file {} does not exist!'.format(filename))

        #: The logger used by fastr, set and updated by the Config object
        self._update_logging()

    def register_fields(self, fields_spec):
        """
        Register extra fields to the configuration manager.
        """
        if not isinstance(fields_spec, dict):
            raise TypeError("The fields_spec argument should be a dict")

        for name, field_spec in fields_spec.items():
            if name not in self._fields:
                self._fields[name] = field_spec
            else:
                if self._fields[name] != field_spec:
                    raise ValueError('Found a conflicting definition for the field {}'.format(name))

        self._create_field_properties(fields_spec)

    def get_field(self, item):
        if item in self._fields:
            # Get the field specification from the field definitions
            field_spec = self._fields[item]

            # Get the currently stored value of the config
            value = self._current_config.get(item, None)

            # Use the default in case the value is not found
            if value is None:
                value = field_spec[1]

            # Catch cases without a valid defautt
            if value is None:
                raise ValueError('No default given for {}'.format(item))

            # In case of dict/list, get data and merge with default
            if isinstance(value, EmptyDefault):
                if field_spec[0] is dict:
                    new_value = dict(field_spec[1])
                    new_value.update(value.asdict())
                    value = new_value
                elif field_spec[0] is list:
                    new_value = []
                    for x in value.aslist():
                        if x is not NotImplemented:
                            new_value.append(x)
                        else:
                            new_value.extend(field_spec[1])
                    value = new_value
                else:
                    raise TypeError('Config value and type do not match!')

            # Validate the type against the field specification
            if not isinstance(value, field_spec[0]):
                raise TypeError('Config value for {} is of wrong type, expected {}, found {}'.format(item,
                                                                                                     field_spec[0],
                                                                                                     type(value).__name__))

            return value
        else:
            raise KeyError('Field not found in config field specification!')

    def set_field(self, item, value):
        if item in self._fields:
            # Get the field specification from the field definitions
            field_spec = self._fields[item]

            if not isinstance(value, field_spec[0]):
                raise TypeError('Type mismatch, config field should be a {} found {}'.format(
                    field_spec[0],
                    type(value).__name__
                ))

            # Set the value
            self._current_config[item] = value
        else:
            raise KeyError('Field not found in config field specification!')

    @classmethod
    def _create_field_properties(cls, fields):
        """
        Create properties for a dictionary of fields

        :param fields: The mapping of the fields to create
        """
        for field_name, field_value in fields.items():
            if not hasattr(cls, field_name):
                prop = cls._field_property(field_name, field_value[2])
                setattr(cls, field_name, prop)

    @staticmethod
    def _field_property(field_name, field_doc=None):
        """
        Create a property for a field
        :param field_name: the name of the field
        :param field_doc: the docstring for the field
        :return: property to use
        """
        def getter(obj):
            return obj.get_field(field_name)

        def setter(obj, value):
            obj.set_field(field_name, value)

        return property(getter, setter, field_doc)

    def __repr__(self):
        items = []

        for key, spec in self._fields.items():
            val = getattr(self, key)

            if isinstance(val, (dict, list, basestring)):
                val = json.dumps(val, indent=2)

            items.append("# [{s[0].__name__}] {s[2]}\n{k} = {v}\n".format(s=spec, k=key, v=val))

        return '\n'.join(items)

    def read_config(self, filename):
        """
        Read a configuration and update the configuration object accordingly

        :param filename: the configuration file to read
        """
        execfile(filename, {}, self._current_config)

        for key, value in self._current_config.items():
            if isinstance(value, (list, dict)):
                self._current_config[key] = EmptyDefault(value)

        self.read_config_files.append(filename)
        self._update_logging()

    def web_url(self):
        """ Construct a fqdn from the web['hostname'] and web['port'] settings.
        :return: FQDN
        :rtype: str
        """
        #if self.web['port'] == 80:
        #    return 'http://{}' .format(self.web['hostname'])
        #elif self.web['port'] == 443:
        #    return 'https://{}' .format(self.web['hostname'])
        #else:
        #    return 'http://{}:{}'.format(self.web['hostname'], self.web['port'])
        return 'http://localhost'

    def _update_logging(self):
        """
        Update the logging using the current config settings.
        """
        # Create log dir if needed (make sure it exists before other
        # operation requiring logging) but after logdir is known
        if not os.path.exists(self.logdir):
            os.makedirs(self.logdir)

        logging_definition = {
            'version': 1,
            'disable_existing_loggers': True,
            'formatters': {
                'verbose': {
                    'format': '[%(processName)s::%(threadName)s] %(levelname)s: %(module)s:%(lineno)d >> %(message)s'
                },
                'console_simple': {
                    'format': '[%(levelname)s] %(module) 9s:%(lineno)04d >> %(message)s'
                },
                'console_minimal': {
                    'format': '%(levelname)s >> %(message)s'
                },
            },
            'handlers': {
                'console': {
                    'level': 'INFO',
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose' if self.debug else 'console_simple',
                    'stream': 'ext://sys.stdout',
                },
                'childprocess': {
                    'level': 'CRITICAL',
                    'class': 'logging.NullHandler',
                },
                'error_file': {
                    'level': 'ERROR',
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'verbose',
                    'filename': os.path.join(self.logdir, 'error.log'),
                    'maxBytes': 10 * 1024 * 1024,
                    'backupCount': 20,
                },
                'info_file': {
                    'level': 'INFO',
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'verbose',
                    'filename': os.path.join(self.logdir, 'info.log'),
                    'maxBytes': 10 * 1024 * 1024,
                    'backupCount': 20,
                },
                'server_file': {
                    'level': 'INFO',
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'verbose',
                    'filename': os.path.join(self.logdir, 'server.log'),
                    'maxBytes': 10 * 1024 * 1024,
                    'backupCount': 20,
                },
                'null_handler': {
                    'level': 'CRITICAL',
                    'class': 'logging.NullHandler',
                }
            },
            'loggers': {
                'fastr': {
                    'handlers': ['console', 'info_file', 'error_file'],
                    'propagate': True,
                    'level': 'DEBUG',
                },
                'py:warnings': {
                    'handlers': ['console', 'info_file', 'error_file'],
                    'propagate': True,
                }
            },
            'root': {
                'handlers': ['null_handler'],
                'level': 'DEBUG'
            }
        }

        fastr_log_type_options = {
            'default': ['console', 'info_file', 'error_file'],
            'server': ['server_file', 'console'],
            'daemon': ['server_file'],
            'console': ['console'],
            'childprocess': ['childprocess'],
            'worker': ['worker'],
            'none': ['null_handler']
        }

        logging_definition['loggers']['fastr']['handlers'] = fastr_log_type_options[self.logtype]

        if self.debug:
            logging_definition['handlers']['console']['level'] = 'DEBUG'
            logging_definition['handlers']['server_file']['level'] = 'DEBUG'
            logging_definition['handlers']['debug_file'] = {
                'level': 'DEBUG',
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'verbose',
                'filename': os.path.join(self.logdir, 'debug.log'),
                'maxBytes': 10 * 1024 * 1024,
                'backupCount': 20,
                }
            logging_definition['handlers']['childprocess'] = {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'verbose',
                'stream': 'ext://sys.stdout',
                }
            logging_definition['loggers']['fastr']['handlers'].append('debug_file')
        else:
            logging_definition['handlers']['console']['level'] = 'INFO'
            logging_definition['handlers']['server_file']['level'] = 'INFO'
            try:
                del logging_definition['handlers']['debug_file']
            except KeyError:
                pass
            try:
                logging_definition['loggers']['fastr']['handlers'].remove('debug_file')
            except ValueError:
                pass
            logging_definition['loggers']['fastr']['level'] = 'INFO'

        logging.config.dictConfig(logging_definition)
        if self.log is None:
            self.log = logging.getLogger('fastr')
            self.log.debug('Setting up the FASTR environment')
        else:
            self.log.debug('Updated fastr logging')
        self.log.debug('Log directory: {}'.format(self.logdir))
        self.log.debug('Using log type: {} (debug: {})'.format(self.logtype, self.debug))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nose.tools as nt
import fastr
import os
import tempfile
import random
import shutil
import string
import fastr.exceptions as exceptions


class TestVFS:

    def setup(self):
        fastr.log.info('Setup')
        """ Setup the test environment for the vfs tests. """
        # We assume there is a tmp mount. (It should be tested).
        # Generate a file in the tmp mount.
        self.dir = os.path.join(fastr.config.mounts['tmp'], '_test')
        if os.path.exists(self.dir):
            fastr.log.warning('Cleaning up existing temporary test directory {}'.format(self.dir))
            shutil.rmtree(self.dir)
        os.makedirs(self.dir)
        fastr.log.info('Using temporary directory {}'.format(self.dir))
        if not os.path.isdir(self.dir):
            fastr.log.critical('Temporary directory not available!')

        self.source_dir = os.path.join(self.dir, 'source')
        os.mkdir(self.source_dir)
        self.destination_dir = os.path.join(self.dir, 'destination')
        os.mkdir(self.destination_dir)

        self.handle, self.absfilename = tempfile.mkstemp(dir=self.source_dir)
        self.filename = os.path.basename(self.absfilename)
        self.niigz_handle, self.niigz_absfilename = tempfile.mkstemp(dir=self.source_dir, suffix='.nii.gz')
        self.niigz_filename = os.path.basename(self.niigz_absfilename)

        fastr.log.info('Created {} and {}'.format(self.absfilename, self.niigz_absfilename))
        if not os.path.exists(self.absfilename):
            fastr.log.critical('Source file {} does not exist!'.format(self.absfilename))
        if not os.path.exists(self.niigz_absfilename):
            fastr.log.critical('Source file {} does not exist!'.format(self.niigz_absfilename))

    def teardown(self):
        """ Tear down the test environment for the vfs tests. """
        # Delete the generated temporary file
        shutil.rmtree(self.dir)

    def random_unique_string(self, length=8, existing=None):
        """ Return a random alpha-numeric string with a certain length. """
        s = ''.join([random.choice(string.ascii_letters + string.digits) for ch in range(length)])
        if existing is not None and s in existing:
            s = self.random_unique_string(length, existing)
        return s

    def test_url_to_path(self):
        nt.eq_(fastr.vfs.url_to_path('vfs://tmp/_test/source/{}'.format(self.filename)), self.absfilename)

    def test_path_to_url(self):
        nt.eq_(fastr.vfs.path_to_url(self.absfilename), 'vfs://tmp/_test/source/{}'.format(self.filename))

    @nt.raises(exceptions.FastrUnknownURLSchemeError)
    def test_url_to_path_unknown_scheme(self):
        # Get a random sequence of strings to get a non existing scheme.
        s = self.random_unique_string(length=5, existing=fastr.ioplugins.keys())
        fastr.vfs.url_to_path("{}://tmp/blaat.nii.gz".format(s))

    @nt.raises(exceptions.FastrMountUnknownError)
    def test_url_to_path_unknown_mount(self):
        # Get a random sequence of strings to get a non existing scheme.
        m = self.random_unique_string(length=5, existing=fastr.config.mounts.keys())
        fastr.vfs.url_to_path("vfs://{}/blaat.nii.gz".format(m))

    @nt.raises(exceptions.FastrMountUnknownError)
    def test_path_to_url_unknown_mount(self):
        # Get a random sequence of strings to get a non existing scheme.
        m = self.random_unique_string(length=5, existing=fastr.config.mounts.keys())
        fastr.vfs.path_to_url("/{}/blaat.nii.gz".format(m))

    def test_ioplugins_pull_source_data_no_extension(self):
        output_file = os.path.join(self.destination_dir, self.filename)
        fastr.vfs.pull_source_data('vfs://tmp/_test/source/{}'.format(self.filename),
                                   self.destination_dir,
                                   'id_0',
                                   datatype=fastr.typelist['NiftiImageFile'])
        nt.ok_(os.path.isfile(output_file), 'expected output at {}'.format(output_file))
        if os.path.isfile(output_file):
            os.remove(output_file)

    def test_ioplugins_pull_source_data_niigz(self):
        fastr.vfs.pull_source_data('vfs://tmp/_test/source/{}'.format(self.filename),
                                   self.destination_dir,
                                   'id_nii_0',
                                   datatype=fastr.typelist['NiftiImageFileCompressed'])
        output_file = os.path.join(self.destination_dir, self.filename)
        nt.ok_(os.path.isfile(output_file))
        if os.path.isfile(output_file):
            os.remove(output_file)

    def test_ioplugins_pull_source_data_mhd(self):
        absfilename_mhd = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite.mhd')
        absfilename_raw = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite.raw')
        filename_mhd = os.path.basename(absfilename_mhd)
        filename_raw = os.path.basename(absfilename_raw)

        destination_path = self.destination_dir
        destination_path_mhd = os.path.join(self.destination_dir, filename_mhd)
        destination_path_raw = os.path.join(self.destination_dir, filename_raw)

        if os.path.exists(destination_path_mhd):
            os.remove(destination_path_mhd)
        if os.path.exists(destination_path_raw):
            os.remove(destination_path_raw)

        fastr.vfs.pull_source_data(fastr.vfs.path_to_url(absfilename_mhd),
                                   destination_path,
                                   'id_mhd_0',
                                   datatype=fastr.typelist['ITKImageFile'])
        new_mhd_exists = os.path.isfile(destination_path_mhd)
        new_raw_exists = os.path.isfile(destination_path_raw)
        nt.ok_(new_mhd_exists, 'expected output .mhd file {} not found'.format(destination_path_mhd))
        nt.ok_(new_raw_exists, 'expected output .raw file {} not found'.format(destination_path_raw))

    def test_ioplugins_push_sink_data_niigz(self):
        random_filename = self.random_unique_string()
        output_file = os.path.join(self.destination_dir, random_filename) + '.nii.gz'
        output_url = fastr.vfs.path_to_url(output_file)
        fastr.vfs.push_sink_data(self.niigz_absfilename, output_url)
        nt.ok_(os.path.isfile(output_file))
        if os.path.isfile(output_file):
            os.remove(output_file)

    def test_ioplugins_push_sink_data_mhd(self):
        absfilename_mhd = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite.mhd')
        absfilename_raw = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite.raw')
        filename_mhd = os.path.basename(absfilename_mhd)
        filename_raw = os.path.basename(absfilename_raw)

        push_target = os.path.join(self.destination_dir, 'sink')
        os.mkdir(push_target)
        destination_path_mhd = os.path.join(push_target, filename_mhd)
        destination_path_raw = os.path.join(push_target, filename_raw)
        destination_url_mhd = fastr.vfs.path_to_url(destination_path_mhd)

        if os.path.exists(destination_path_mhd):
            os.remove(destination_path_mhd)
        if os.path.exists(destination_path_raw):
            os.remove(destination_path_raw)

        fastr.vfs.push_sink_data(absfilename_mhd, destination_url_mhd)
        new_mhd_exists = os.path.isfile(destination_path_mhd)
        new_raw_exists = os.path.isfile(destination_path_raw)
        nt.ok_(new_mhd_exists, 'expected output .mhd file {} not found'.format(destination_path_mhd))
        nt.ok_(new_raw_exists, 'expected output .raw file {} not found'.format(destination_path_raw))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains a class to interact with EnvironmentModules
"""

import os
import re
import subprocess

from enum import Enum

import fastr
import fastr.exceptions as exceptions
from fastr.utils.procutils import which


class ModuleSystem(Enum):
    lmod = 'Lmod'
    envmod = 'enviromentmodules'


class EnvironmentModules(object):
    """
    This class can control the module environments in python. It can list, load
    and unload environmentmodules. These modules are then used if subprocess is
     called from python.
    """

    try:
        # Do the initialization of Environment Modules of the system
        if 'MODULESHOME' not in os.environ:
            raise exceptions.FastrImportError('Cannot find Environment Modules home directory (environment variables not setup properly?)')

        if 'MODULE_VERSION' not in os.environ:
            os.environ['MODULE_VERSION_STACK'] = '3.2.6'
            os.environ['MODULE_VERSION'] = '3.2.6'
        else:
            os.environ['MODULE_VERSION_STACK'] = os.environ['MODULE_VERSION']

        if 'MODULEPATH' not in os.environ:
            try:
                __paths = []
                with open(os.path.join(os.environ['MODULESHOME'], 'init', '.modulespath'), 'r') as fin:
                    for line in fin:
                        line = re.sub(r'(?P<path>[^#]*)(\#.*)', '\\g<path>', line).strip()
                        if line:
                            __paths.append(line)
            except:
                raise exceptions.FastrImportError('Cannot load Environment Modules path defintion!')

            os.environ['MODULEPATH'] = ':'.join(__paths)

        if 'LOADEDMODULES' not in os.environ:
            os.environ['LOADEDMODULES'] = ''

        if 'LMOD_CMD' in os.environ:
            # Lua mod is installed, we prefer lua mod
            _bin = os.environ['LMOD_CMD']
            _module_settings_system = ModuleSystem.lmod
        else:
            # The old environment modules is installed
            _bin = '{}/bin/modulecmd'.format(os.environ['MODULESHOME'])
            _module_settings_system = ModuleSystem.envmod

        if not os.path.isfile(_bin) or not os.access(_bin, os.X_OK):
            _bin = which('modulecmd')
            if _bin is None:
                raise exceptions.FastrImportError('Cannot find Environment Modules executable!')

    except exceptions.FastrImportError as exception:
        _module_settings_loaded = False
        _module_settings_warning = exception.message
    else:
        _module_settings_loaded = True
        _module_settings_warning = ''

    def __init__(self, protected=None):
        """
        Create the environmentmodules control object

        :param list protected: list of modules that should never be unloaded
        :return: newly created EnvironmentModules
        """
        if not self._module_settings_loaded:
            raise exceptions.FastrValueError('Could not load the environmentmodules settings during import, this module is non-functional! Original exception: {}'.format(self._module_settings_warning))

        # Debug print the system used
        fastr.log.debug('Using the {} system'.format(self._module_settings_system))

        # Set initial versions
        self._avail_string = ''
        self._avail_modules = ()
        self._loaded_modules = ()

        # Sync all available and loaded Modules
        self.sync()

        if protected is not None:
            self._protected = protected
        else:
            self._protected = []

    def __repr__(self):
        return "<EnvironmentModules system={}>".format(self._module_settings_system.value)

    def sync(self):
        """
        Sync the object with the underlying enviroment. Re-checks the
        available and loaded modules
        """
        self._sync_loaded()
        self._sync_avail()

    def _sync_loaded(self):
        """
        Sync the loaded modules with the underlying environmentmodules
        """
        _, stderr = self._module('list')

        loaded_modules = []
        for line in stderr.splitlines():
            for module in re.findall(r'\W*(\d+)\) ([^/^\(^\)^\W]+)(/[^\(^\)^ ^\t]+)?', line):
                loaded_modules.append(module)

        # Sort and remove order from tuples
        loaded_modules.sort()
        loaded_modules = [module[1:] for module in loaded_modules]

        self._loaded_modules = tuple(loaded_modules)

    def _sync_avail(self):
        """
        Sync the available modules with the underlying environmentmodules
        """
        _, self._avail_string = self._module('avail')

        modules = []
        for line in self._avail_string.split('\n'):
            # Skip padding lines etc
            if len(line) == 0 or line[0] == '-' or '/' not in line:
                continue

            # Parse the line
            for part in line.split():
                if part == '(L)':
                    continue

                if part == '(D)':
                    modules[-1] = modules[-1][:-1] + (True,)
                    continue

                if part.endswith('(default)'):
                    default = True
                    part = part[:-9]
                else:
                    default = False

                modules.append(tuple(part.split('/') + [default]))

        # Sort modules by name/version
        modules.sort()

        self._avail_modules = tuple(modules)

    def _module(self, *arguments):
        """
        Call the environmentmodule executable

        :param arguments: arguments similar to the commandline version
        :return: tuple with (stdout, stderr)
        """
        args = [self._bin, 'python']
        args.extend(arguments)
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()

        return (stdout, stderr)

    @staticmethod
    def totuple_modvalue(value):
        """
        Turn a representation of a module into a tuple representation

        :param value: module representation (either str or tuple)
        :return: tuple representation (name, version, default)
        """
        if isinstance(value, str):
            match = re.match(r'([^/^\(^\)]+)([^\(^\)^ ^\t]+)?(\(default\))?', value)
            if match:
                value = match.groups()
                return value[:-1]
            else:
                raise exceptions.FastrValueError('Invalid module string ({})!'.format(value))
        elif isinstance(value, tuple):
            if len(value) in [2, 3]:
                return value
            else:
                raise exceptions.FastrValueError('Invalid module tuple ({})!'.format(value))
        else:
            raise exceptions.FastrTypeError('Invalid module value type ({})!'.format(type(value.__name__)))

    @staticmethod
    def tostring_modvalue(value):
        """
        Turn a  representation of a module into a string representation

        :param value: module representation (either str or tuple)
        :return: string representation
        """
        if isinstance(value, str):
            return value
        elif isinstance(value, tuple):
            return ''.join(value)
        else:
            raise exceptions.FastrTypeError('Invalid module value type ({})!'.format(type(value.__name__)))

    def _run_commands_string(self, value):
        """
        Run the commands returned by environmentmodules, and warn for
        non-default commands

        :param value: str of commands returned by environmentmodules
        """
        for command in value.splitlines():
            match1 = re.match(r'^os.environ\[(?P<sep1>[\'"])(\w+)(?P=sep1)\] = (?P<sep2>[\'"]).*(?P=sep2);?\s*$', command)
            if match1 is None:
                match2 = re.match(r'^del os.environ\[(?P<sep1>[\'"])(\w+)(?P=sep1)\];?\s*$', command)
                if match2 is None:
                    fastr.log.warning('WARNING: Evaluating unchecked code: {!r}'.format(command))
            # We know this is bad practice, but we trust the code from environmentmodules
            # pylint: disable=exec-used
            exec command

        # Make sure loaded module list is synced
        self._sync_loaded()

    @property
    def loaded_modules(self):
        """
        List of currently loaded modules
        """
        return [''.join(x) for x in self._loaded_modules]

    @property
    def avail_modules(self):
        """
        List of avaible modules
        """
        return self._avail_modules

    def avail(self, namestart=None):
        """
        Print available modules in same way as commandline version

        :param namestart: filter on modules that start with namestart
        """
        for module in self._avail_modules:
            default = module[-1]
            module = '/'.join(module[:-1])
            if namestart is None or module.startswith(namestart):
                if default:
                    fastr.log.info('{} (default)'.format(module))
                else:
                    fastr.log.info(module)


    def isloaded(self, module):
        """
        Check if a specific module is loaded

        :param module: module to check
        :return: flag indicating the module is loaded
        """
        module = self.totuple_modvalue(module)

        if module[1] is not None:
            return module in self._loaded_modules
        else:
            for mod in self._loaded_modules:
                if module[0] == mod[0]:
                    return True
            return False

    def load(self, module):
        """
        Load specified module

        :param module: module to load
        """
        module = self.tostring_modvalue(module)
        if self.isloaded(module):
            fastr.log.warning('Module {} is already loaded, unloading first!'.format(module))
            self.unload(module)

        stdout, stderr = self._module('load', module)

        if len(stderr) != 0:
            fastr.log.error('Encountered error when loading module:\n{}'.format(stderr))
        else:
            self._run_commands_string(stdout)

    def unload(self, module):
        """
        Unload specified module

        :param module: module to unload
        """
        module = self.tostring_modvalue(module)

        if module in self._protected:
            fastr.log.debug('Module {} is protected, skipping unload!'.format(module))
            return

        if not self.isloaded(module):
            fastr.log.warning('Module {} is not loaded, cannot unload!'.format(module))

        stdout, stderr = self._module('unload', module)

        if len(stderr) != 0:
            fastr.log.error('Encountered error when unloading module:\n{}'.format(stderr))
        else:
            self._run_commands_string(stdout)

    def reload(self, module):
        """
        Reload specified module

        :param module: module to reload
        """
        module = self.tostring_modvalue(module)
        if not self.isloaded(module):
            fastr.log.warning(': Module {} is not loaded, cannot unload first!'.format(module))
        else:
            stdout, stderr = self._module('unload', module)
            if len(stderr) != 0:
                fastr.log.error('Encountered error when unloading module:\n{}'.format(stderr))
            else:
                self._run_commands_string(stdout)

        stdout, stderr = self._module('load', module)
        if len(stderr) != 0:
            fastr.log.error('Encountered error when loading module:\n{}'.format(stderr))
        else:
            self._run_commands_string(stdout)

    def swap(self, module1, module2):
        """
        Swap one module for another one

        :param module1: module to unload
        :param module2: module to load
        """
        self.unload(module1)
        self.load(module2)

    def clear(self):
        """
        Unload all modules (except the protected modules as they cannot be
        unloaded). This should result in a clean environment.
        """
        for module in self.loaded_modules:
            self.unload(module)



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastr

IS_TEST = False  # Do not use as a test


def create_network():
    network = fastr.Network(id_="test_elastix_nipype_demo")

    source1 = network.create_source('ITKImageFile', id_='fixed_img')
    source2 = network.create_source('ITKImageFile', id_='moving_img')
    param1 = network.create_source('ElastixParameterFile', id_='param_file')

    elastix_node = network.create_node('NipypeElastix', id_='elastix')
    elastix_node.inputs['fixed_image'] = source1.output
    elastix_node.inputs['moving_image'] = source2.output
    link_param = network.create_link(param1.output, elastix_node.inputs['parameters'])
    link_param.converge = 0

    outtrans = network.create_sink('ElastixTransformFile', id_='sink_trans')
    outtrans.inputs['input'] = elastix_node.outputs['transform']

    transformix_node = network.create_node('Transformix', id_='transformix')
    transformix_node.inputs['image'] = source2.output
    transformix_node.inputs['transform'] = elastix_node.outputs['transform'][-1]

    outimage = network.create_sink('ITKImageFile', id_='sink_image')
    outimage.inputs['input'] = transformix_node.outputs['image']

    network.draw_network(img_format='svg')
    network.dumpf('{}.json'.format(network.id), indent=2)

    return network


def source_data(network):
    return {'fixed_img': {'s1': 'vfs://example_tools/elastix_test/img0/slice047.mhd'},
            'moving_img': {'s1': 'vfs://example_tools/elastix_test/img1/slice091.mhd'},
            'param_file': ['vfs://example_tools/elastix_test/parAslice.txt',
                           'vfs://example_tools/elastix_test/parBslice.txt']}


def sink_data(network):
    return {'sink_trans': 'vfs://tmp/results/elastix_output_trans_{sample_id}_{cardinality}.txt',
            'sink_image': 'vfs://tmp/results/elastix_output_image_{sample_id}.nii.gz'}


def main():
    network = create_network()

    # Execute
    network.draw_network()
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging


def runapp(debug=False):
    from fastr.web import app
    from fastr import config

    # Add werkzeug logger to the mix for debugging http requests.
    if debug:
        _logger = logging.getLogger('werkzeug')
        _logger.setLevel(logging.INFO)
        _logger.addHandler(logging.StreamHandler())

    app.logger.info("Starting web server (on {}:{}).".format(config.web_hostname, config.web_port))
    app.run(host=config.web_hostname, port=config.web_port, debug=debug)
    app.logger.info("Web server has stopped.")


def main():
    parser = argparse.ArgumentParser(description="Fastr web client")
    parser.add_argument('-d', '--debug', action="store_true", help="Debug mode.")
    args = parser.parse_args()

    runapp(args.debug)

if __name__ == "__main__":
    main()



import requests
import fastr

from fastr.core.node import MacroNode
from fastr.execution.job import JobState
from fastr import exceptions


class PimPublisher(object):
    """
    Class to publish to PIM
    """
    PIM_STATUS_MAPPING = {
        JobState.nonexistent: 'unknown',
        JobState.created: 'idle',
        JobState.queued: 'idle',
        JobState.hold: 'idle',
        JobState.running: 'running',
        JobState.execution_done: 'running',
        JobState.execution_failed: 'running',
        JobState.processing_callback: 'running',
        JobState.finished: 'success',
        JobState.failed: 'failed',
        JobState.cancelled: 'failed',
    }

    def __init__(self, uri=None):
        if uri is None and fastr.config.pim_host == '':
            fastr.log.info("No valid PIM host given, PIM publishing will be disabled!")
            self.pim_uri = None
        else:
            self.pim_uri = uri or fastr.config.pim_host
        self.registered = False
        self.run_id = None

    def pim_update_status(self, job):
        if not self.registered:
            fastr.log.debug('Did not register a RUN with PIM yet! Cannot'
                            ' send status updates!')
            return

        run_id = job.run_id

        # Create PIM job data
        pim_job_data = {
            "id": job.id,
            "node_id": job.node_id,
            "run_id": run_id,
            "sample_id": str(job.sample_id),
            "status": self.PIM_STATUS_MAPPING[job.status]
        }

        # Send the data to PIM
        fastr.log.debug('Updating PIM job status {} => {} ({})'.format(job.id,
                                                                       job.status,
                                                                       self.PIM_STATUS_MAPPING[job.status]))
        uri = '{pim}/api/runs/{run_id}/jobs/{job_id}'.format(pim=fastr.config.pim_host,
                                                             run_id=run_id,
                                                             job_id=job.id)

        fastr.log.debug('Send PUT to pim at {}:\n{}'.format(uri, pim_job_data))
        try:
            response = requests.put(uri, json=pim_job_data)
        except requests.ConnectionError as exception:
            fastr.log.error('Could no publish status to PIM, encountered exception: {}'.format(exception))

    @staticmethod
    def pim_serialize_network(network):
        """
        Serialize Network in the correct for to use with PIM.

        :return: json data for PIM
        """
        node_classes = {'Node': 'node',
                        'SourceNode': 'source',
                        'ConstantNode': 'constant',
                        'SinkNode': 'sink'}

        network_data = {
            "description": network.description,
            "nodes": [],
            "links": [],
            "groups": [{"id": 'step_' + x,
                        "description": "undefined",
                        "parent_group": "root"} for x in network.stepids.keys()]
        }

        # Add the nodes
        for node in network.nodelist.values():
            if isinstance(node, MacroNode):
                # MacroNodes are a weird tool-less Node that will fail
                continue

            node_class = node.__class__.__name__
            step = None
            for stepid, nodes in network.stepids.items():
                if node in nodes:
                    step = stepid
                    break

            node_data = {
                "group_id": ('step_' + step) if step else "root",
                "id": node.id,
                "in_ports": [{'id': 'in_' + x.id, 'description': x.description} for x in node.tool.inputs.values()],
                "out_ports": [{'id': 'out_' + x.id, 'description': x.description} for x in node.tool.outputs.values()],
                "type": node_classes[node_class] if node_class in node_classes else 'node'
            }

            network_data["nodes"].append(node_data)

        # Add the links
        for link in network.linklist.values():
            link_data = {
                "id": link.id,
                "from_node": link.source.node.id,
                "from_port": 'out_' + link.source.id,
                "to_node": link.target.node.id,
                "to_port": 'in_' + link.target.id,
                "type": link.source.resulting_datatype.id
            }

            network_data["links"].append(link_data)

        return network_data

    def pim_register_run(self, network):
        if self.pim_uri is None:
            fastr.log.debug('No valid PIM uri known. Cannot register to PIM!')
            return

        self.run_id = network.run_id
        pim_run_data = {
            "collapse": False,
            "description": "Run of {} started at {}".format(network.id,
                                                            network.timestamp),
            "id": self.run_id,
            "network": self.pim_serialize_network(network),
            "workflow_engine": "fastr"
        }

        uri = '{pim}/api/runs/'.format(pim=fastr.config.pim_host)
        fastr.log.info('Registering {} with PIM at {}'.format(self.run_id, uri))

        fastr.log.debug('Send PUT to pim at {}:\n{}'.format(uri, pim_run_data))

        # Send out the response and record if we registered correctly
        try:
            response = requests.put(uri, json=pim_run_data)
            if response.status_code in [200, 201]:
                self.registered = True
        except requests.ConnectionError as exception:
            fastr.log.error('Could no register network to PIM, encountered'
                            ' exception: {}'.format(exception))



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time
import shutil
import random


try:
    if len(sys.argv) == 1:
        sys.argv.append('-h')

    if sys.argv[1] == '-h' or len(sys.argv) < 3:
        print('delays the execution for a random time (somewhere between 2 and 5 seconds)')
        print('usage:')
        print('    ' + sys.argv[0] + ' --in inputfile --out outputfile')
        sys.exit()

    in_index = sys.argv.index('--in') + 1
    out_index = sys.argv.index('--out') + 1

    infile = sys.argv[in_index]
    outfile = sys.argv[out_index]

    rnd = random.randint(2, 5)
    print "delay for {} seconds".format(rnd)
    time.sleep(rnd)

    shutil.copy(infile, outfile)

except KeyboardInterrupt:
    print('   ')
    print('ctrl-c detected, now exiting ... ')


import os
import importlib
import inspect
import sys
import textwrap

from fastr import exceptions


def find_commands():
    cmd_dir = os.path.dirname(__file__)
    files = os.listdir(cmd_dir)
    files = [x[:-3] for x in files if x.endswith('.py') and not x.startswith('_')]

    return files


def get_command_module(command):
    command_module = importlib.import_module('fastr.utils.cmd.{}'.format(command))

    if not hasattr(command_module, 'get_parser') or not inspect.isfunction(getattr(command_module, 'get_parser')):
        raise exceptions.FastrAttributeError('Command modules should contain a "get_parser" function')

    if not hasattr(command_module, 'main') or not inspect.isfunction(getattr(command_module, 'main')):
        raise exceptions.FastrAttributeError('Command modules should contain a "main" function')

    return command_module


def print_help(commands=None):
    if commands is None:
        commands = find_commands()

    # Get length of longest command name
    max_name_length = max(len(x) for x in commands) + 4
    max_text_length = 80 - max_name_length

    help = []
    for command in sorted(commands):
        module = get_command_module(command)
        if hasattr(module.main, '__doc__'):
            doc_string = module.main.__doc__.strip().splitlines()[0]
            doc_string = textwrap.wrap(doc_string, width=max_text_length)
        else:
            doc_string = ['Undocumented territory']

        doc_string = ['{command:<{len}} {text}'.format(command=command if nr == 0 else '',
                                                       len=max_name_length,
                                                       text=text) for nr, text in enumerate(doc_string)]

        help.append('\n'.join(doc_string))

    command_help_list = '\n'.join(help)
    print ("""usage: fastr [-h] subcommand

subcommand:
  The sub command to use


List of sub commands:
{}""".format(command_help_list))


def main():
    # Get all files in utils following the bin_*.py name convention
    commands = find_commands()

    # Strip the first argument, this should be a sub command
    # and if it is not we will display the main help

    if len(sys.argv) < 2:
        print_help()
        return

    subcommand = sys.argv[1]

    if subcommand not in commands:
        print_help()
        return

    # Path sys.argv to look like "fastr subcommand" was one command
    sys.argv = ['{} {}'.format(sys.argv[0], subcommand)] + sys.argv[2:]

    module = get_command_module(subcommand)
    module.main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the CommaSeperatedValueFile plugin for fastr
"""

import csv
import os
import urlparse
import fastr
import fastr.exceptions as exceptions
from fastr.core.ioplugin import IOPlugin


class CommaSeperatedValueFile(IOPlugin):
    """
    The CommaSeperatedValueFile an expand-only type of IOPlugin. No URLs
    can actually be fetched, but it can expand a single URL into a larger
    amount of URLs.

    The ``csv://`` URL is a ``vfs://`` URL with a number of query variables
    available. The URL mount and path should point to a valid CSV file. The
    query variable then specify what column(s) of the file should be used.

    The following variable can be set in the query:

    ============= =============================================================================================
    variable      usage
    ============= =============================================================================================
    value         the column containing the value of interest, can be int for index or string for key
    id            the column containing the sample id (optional)
    header        indicates if the first row is considered the header, can be ``true`` or ``false`` (optional)
    delimiter     the delimiter used in the csv file (optional)
    quote         the quote character used in the csv file (optional)
    reformat      a reformatting string so that ``value = reformat.format(value)`` (used before relative_path)
    relative_path indicates the entries are relative paths (for files), can be ``true`` or ``false`` (optional)
    ============= =============================================================================================

    The header is by default ``false`` if the neither the ``value`` and ``id``
    are set as a string. If either of these are a string, the header is
    required to define the column names and it automatically is assumed ``true``

    The delimiter and quota characters of the file should be detected
    automatically using the :class:`Sniffer <csv.Sniffer>`, but can be forced
    by setting them in the URL.

    Example of valid ``csv`` URLs::

        # Use the first column in the file (no header row assumed)
        csv://mount/some/dir/file.csv?value=0

        # Use the images column in the file (first row is assumed header row)
        csv://mount/some/dir/file.csv?value=images

        # Use the segmentations column in the file (first row is assumed header row)
        # and use the id column as the sample id
        csv://mount/some/dir/file.csv?value=segmentations&id=id

        # Use the first column as the id and the second column as the value
        # and skip the first row (considered the header)
        csv://mount/some/dir/file.csv?value=1&id=0&header=true

        # Use the first column and force the delimiter to be a comma
        csv://mount/some/dir/file.csv?value=0&delimiter=,
    """
    scheme = 'csv'

    def __init__(self):
        super(CommaSeperatedValueFile, self).__init__()

    def expand_url(self, url):
        if fastr.data.url.get_url_scheme(url) != 'csv':
            raise exceptions.FastrValueError('URL not of csv type!')

        # Retrieve base path
        parsed = urlparse.urlparse(url)
        csvurl = urlparse.urlunparse(urlparse.ParseResult(scheme='vfs', netloc=parsed.netloc, path=parsed.path, params='', query='', fragment=''))
        baseurl = urlparse.urlunparse(urlparse.ParseResult(scheme='vfs', netloc=parsed.netloc, path=os.path.dirname(parsed.path), params='', query='', fragment=''))
        csvpath = fastr.ioplugins.url_to_path(csvurl)
        query = urlparse.parse_qs(parsed.query)

        # Get dialect options
        dialect_override = {}
        if 'delimiter' in query:
            dialect_override['delimiter'] = query['delimiter'][0]
        if 'quote' in query:
            dialect_override['quotechar'] = query['quote'][0]

        # Check which column the values are in
        data_column = query['value'][0]
        try:
            data_column = int(data_column)
        except ValueError:
            pass

        # Check which column the ids are in
        id_column = query.get('id', [None])[0]
        try:
            id_column = int(id_column)
        except (ValueError, TypeError):
            pass

        # Get info about hear row
        header_row = {'0': False, '1': True, 'false': False, 'true': True, 'none': None}[query.get('header', ('none',))[0]]

        if header_row is None:
            header_row = isinstance(data_column, str) or isinstance(id_column, str)
        elif not header_row:
            if isinstance(data_column, str) or isinstance(id_column, str):
                raise ValueError('Cannot use string keys if there is no header row')

        # Check if we expect relative paths
        relpath = {'0': False, '1': True, 'false': False, 'true': True, 'none': False}[query.get('relative_path', ('none',))[0]]
        if baseurl[-1] != '/':
            baseurl = baseurl + '/'

        reformat = query.get('reformat', (None,))[0]

        with open(csvpath, 'rb') as csvfile:
            dialect = csv.Sniffer().sniff(csvfile.read(4096))
            csvfile.seek(0)
            reader = csv.reader(csvfile, dialect, **dialect_override)

            values = []
            ids = []
            for row in reader:
                if header_row:
                    if isinstance(id_column, str):
                        id_column = row.index(id_column)
                    if isinstance(data_column, str):
                        data_column = row.index(data_column)

                    header_row = False
                    continue

                current_value = row[data_column]

                if reformat is not None:
                    current_value = reformat.format(current_value)

                if relpath:
                    current_value = baseurl + current_value

                values.append(current_value)

                ids.append(row[id_column] if id_column is not None else None)

        return tuple((i, value) for i, value in zip(ids, values))





# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package holds the classes for working with samples.
"""

from abc import ABCMeta, abstractmethod, abstractproperty
import itertools
from collections import MutableMapping, OrderedDict
import fastr
import fastr.datatypes
import fastr.exceptions as exceptions


class SampleItemBase(tuple):
    """
    This class represents a sample item, a combination of a SampleIndex,
    SampleID, value and required jobs. The SampleItem based on a named
    tuple and has some extra methods to combine SampleItems easily.
    """

    def __new__(cls, index, id_, data, jobs=None, failed_annotations=None):
        """
        Create a SampleItem. Data should be an OrderedDict of tuples.

        :param index: the sample index
        :type index: tuple, slice
        :param id_: the sample id
        :type id_: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param data: the data values
        :type data: SampleValue, Mapping
        :param set jobs: set, tuple or list of jobs on which this SampleItems data depends.
        :param set failed_annotations: set of tuples. The tuple is contructed like follows: (job_id, reason).
        """

        if not isinstance(index, SampleIndex):
            index = SampleIndex(index)

        if not isinstance(id_, SampleId):
            id_ = SampleId(id_)

        if jobs is None:
            jobs = set()

        if not isinstance(jobs, (set, tuple, list)):
            raise exceptions.FastrTypeError("Jobs should be a set, tuple or list of jobs or job ids, found {}".format(type(jobs).__name__))

        jobs = {job if isinstance(job, str) else job.id for job in jobs}

        if failed_annotations is None:
            failed_annotations = set()

        if not isinstance(failed_annotations, set):
            raise exceptions.FastrTypeError("failed_annotations should be a set, found {}".format(type(failed_annotations).__name__))

        return super(SampleItemBase, cls).__new__(cls, [index, id_, data, jobs, failed_annotations])

    def __repr__(self):
        """
        Get a string representation for the SampleItem

        :return: the string representation
        :rtype: str
        """
        return '<{cls} index={s.index}, id={s.id}>'.format(cls=type(self).__name__, s=self)

    def __getnewargs__(self):
        """
        Get new args gives the arguments to use to re-create this object, This
        is used for serialization.
        """
        return tuple(self)

    def __add__(self, other):
        """
        The addition operator combines two SampleItems into a single
        SampleItems. It merges the data and jobs and takes the index
        and id of the left-hand item.

        :param other: The other item to add to this one
        :type other: SampleItem
        :return: the combined SampleItem
        :rtype: SampleItem
        """
        if not isinstance(other, SampleItemBase):
            return NotImplemented

        return type(self)(self.index,
                          self.id,
                          self.data + other.data,
                          self.jobs | other.jobs,
                          self.failed_annotations | other.failed_annotations)

    @staticmethod
    def combine(*args):
        """
        Combine a number of SampleItems into a new one.

        :param *args: the SampleItems to combine
        :type args: iterable of SampleItems
        :return: the combined SampleItem
        :rtype: SampleItem

        It is possible to both give multiple arguments, where each argument is
        a SampleItem, or a single argument which is an iterable yielding
        SampleItems.

        .. code-block:: python

          # variables a, b, c, d are SampleItems to combine
          # These are all valid ways of combining the SampleItems
          comb1 = SampleItem.combine(a, b, c, d)  # Using multiple arguments
          l = [a, b, c, d]
          comb2 = SampleItem.combine(l)  # Using a list of arguments
          comb3 = SampleItem.combine(l.__iter__())  # Using an iterator
        """
        if len(args) == 1 and not isinstance(args[0], SampleItemBase):
            args = args[0]

        args = [x for x in args]

        if not all(isinstance(x, SampleItemBase) for x in args):
            raise exceptions.FastrTypeError("All arguments should be SampleItems! Found {}".format(args))

        if len(args) == 0:
            raise exceptions.FastrValueError("Cannot combine empty list")
        elif len(args) == 1:
            return args[0]
        else:
            result = args[0]
            for item in args[1:]:
                result = result + item
            return result

    @property
    def index(self):
        """
        The index of the SampleItem

        :return: The index of this SampleItem
        :rtype: SampleIndex
        """
        return self[0]

    @property
    def id(self):
        """
        The sample id of the SampleItem

        :return: The id of this SampleItem
        :rtype: SampleId
        """
        return self[1]

    @property
    def data(self):
        """
        The data SampleValue of the SampleItem

        :return: The value of this SampleItem
        :rtype: SampleValue
        """
        return self[2]

    @property
    def jobs(self):
        """
        The set of the jobs on which this SampleItem depends

        :return: The jobs that generated the data for this SampleItem
        :rtype: set
        """
        return self[3]

    @property
    def failed_annotations(self):
        return self[4]

    @property
    def cardinality(self):
        """
        The cardinality of this Sample
        """
        return len(self.data)

    @property
    def dimensionality(self):
        """
        The dimensionality of this Sample
        """
        return len(self.index)


class SampleItem(SampleItemBase):
    def __new__(cls, index, id_, data, jobs=None, failed_annotations=None):
        """
        Create a SampleItem. Data should be an OrderedDict of tuples.

        :param index: the sample index
        :type index: tuple, slice
        :param id_: the sample id
        :type id_: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param data: the data values
        :type data: SampleValue, Mapping
        :param set jobs: set of jobs on which this SampleItems data depends.
        :param set failed_annotations: set of tuples. The tuple is contructed like follows: (job_id, reason).
        """

        if not isinstance(data, SampleValue):
            data = SampleValue(data)

        if not all(isinstance(x, (int, str)) for x in data.keys()):
            types = set(type(x).__name__ for x in data.keys())
            raise exceptions.FastrTypeError("All keys in data should be tuples, found {}".format(types))

        if not all(isinstance(x, tuple) for x in data.values()):
            types = set(type(x).__name__ for x in data.values())
            raise exceptions.FastrTypeError("All values in data should be tuples, found {}".format(types))

        return super(SampleItem, cls).__new__(cls, index, id_, data, jobs, failed_annotations)


class SamplePayload(SampleItemBase):
    def __new__(cls, index, id_, data, jobs=None, failed_annotations=None):
        """
        Create a SampleItem. Data should be an OrderedDict of tuples.

        :param index: the sample index
        :type index: tuple, slice
        :param id_: the sample id
        :type id_: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param data: the data values
        :type data: SampleValue, Mapping
        :param set jobs: set of jobs on which this SampleItems data depends.
        :param set failed_annotations: set of tuples. The tuple is contructed like follows: (job_id, reason).
        """

        if not isinstance(data, MutableMapping):
            raise exceptions.FastrTypeError('SamplePayload should have a MutableMapping as data')

        if not all(isinstance(x, str) for x in data.keys()):
            types = set(type(x).__name__ for x in data.keys())
            raise exceptions.FastrTypeError("All keys in data should be str, found {}".format(types))

        if not all(isinstance(x, SampleItem) for x in data.values()):
            types = set(type(x).__name__ for x in data.values())
            raise exceptions.FastrTypeError("All values in data should be SampleItems, found {}".format(types))

        return super(SamplePayload, cls).__new__(cls, index, id_, data, jobs, failed_annotations)

    def __add__(self, other):
        """
        The addition operator combines two SampleItems into a single
        SampleItems. It merges the data and jobs and takes the index
        and id of the left-hand item.

        :param other: The other item to add to this one
        :type other: SampleItem
        :return: the combined SamplePayload
        :rtype: SamplePayload
        """
        if not isinstance(other, SamplePayload):
            return NotImplemented

        common_keys = set(self.data.keys()) & set(other.data.keys())

        if len(common_keys) > 0:
            raise exceptions.FastrValueError('Cannot combine SamplePayloads, they both contain definitions for {}'.format(common_keys))

        new_data = OrderedDict(self.data)
        new_data.update(other.data)
        return SamplePayload(self.index,
                             self.id,
                             new_data,
                             self.jobs | other.jobs,
                             self.failed_annotations | other.failed_annotations)


class HasSamples(object):
    """
    Base class for all classes that supply samples. This base class allows
    to only define __getitem__ and size and get all other basic functions mixed
    in so that the object behaves similar to a Mapping.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError()

    @abstractproperty
    def size(self):
        raise NotImplementedError()

    def __contains__(self, item):
        try:
            self[item]
        except (KeyError, IndexError):
            return False
        else:
            return True

    def iteritems(self):
        if len(self.size) > 0:
            if None in self.size:
                raise exceptions.FastrValueError('The size {} contains None, cannot continue!')
            for index in itertools.product(*[range(x) for x in self.size]):
                try:
                    yield self[SampleIndex(index)]
                except exceptions.FastrIndexNonexistent:
                    pass

    def __iter__(self):
        for item in self.iteritems():
            yield item.index

    def items(self):
        return list(self.iteritems())

    def indexes(self):
        return list(self)

    def ids(self):
        return [self[x].id for x in self.iteritems()]


class SampleBaseId(tuple):
    """
    This class represents a sample id. A sample id is a multi-dimensional
    id that has a simple, consistent string representation.
    """
    _element_type = type(None)

    def __new__(cls, *args):
        """
        Create a new SampleId

        :param args: the strings to make sample id for
        :type args: iterator/iterable of element type or element type
        """
        # Probably an iterable or str, so only use first element
        if len(args) == 1:
            args = args[0]

        # Make sure we won't iterate over character in a str
        if isinstance(args, cls._element_type):
            args = (args,)

        # Use the generator to create values and store as tuple for later use
        try:
            args = [x for x in args]
        except TypeError:
            if isinstance(cls._element_type, tuple):
                etype = '/'.join(x.__name__ for x in cls._element_type)
            else:
                etype = cls._element_type.__name__
            raise exceptions.FastrTypeError('The arguments for {} should be a iterable of {} elements, found non-iterable {}'.format(cls.__name__, etype, args))

        # Make sure the ID is not empty (which would make a terrible identifier)
        if len(args) == 0:
            raise exceptions.FastrValueError('A {} must have at least one dimension to be valid!'.format(cls.__name__))

        if not all(isinstance(x, cls._element_type) for x in args):
            raise exceptions.FastrTypeError(
                'The arguments for {} should be a iterable of {} elements, found iterable {}'.format(cls.__name__,
                                                                                                     cls._element_type,
                                                                                                     args)
            )

        return super(SampleBaseId, cls).__new__(cls, args)

    def __repr__(self):
        """
        Get a string representation for the SampleBaseId

        :return: the string representation
        :rtype: str
        """
        return '<{} {}>'.format(type(self).__name__, super(SampleBaseId, self).__repr__())

    def __str__(self):
        """
        Get a string version for the SampleId, joins the SampleId with __ to
        create a single string version.

        :return: the string version
        :rtype: str
        """
        return '__'.join(str(x) for x in self)

    def __add__(self, other):
        """
        Add another SampleId, this allows to add parts to the SampleId in a
        convenient way.
        """
        if isinstance(other, self._element_type):
            other = (other,)

        if not isinstance(other, tuple):
            return NotImplemented

        return type(self)(super(SampleBaseId, self).__add__(other))

    def __radd__(self, other):
        """
        Add another SampleId, this allows to add parts to the SampleId in a
        convenient way. This is the right-hand version of the operator.
        """

        if isinstance(other, self._element_type):
            other = (other,)

        if not isinstance(other, tuple):
            return NotImplemented

        return type(self)(tuple(other) + tuple(self))


class SampleId(SampleBaseId):
    """
    SampleId is an identifier for data using human readable strings
    """
    _element_type = (str, unicode)


class SampleIndex(SampleBaseId):
    """
    SampleId is an identifier for data using the location in the N-d data structure.
    """
    _element_type = (int, slice)

    def __str__(self):
        """
        Get a string version for the SampleId, joins the SampleId with __ to
        create a single string version.

        :return: the string version
        :rtype: str
        """
        # This horrible one-liner returns the str representation of the elements
        # For an int is just calls str, but for a slice it will create a str like
        # start:stop[:step] with the None as empty
        plot_elem = [('{x[0]}:{x[1]}' if s.step is None else '{x[0]}:{x[1]}:{x[2]}').format(
            x=[str(n) if n is not None else '' for n in (s.start, s.stop, s.step)])
            if isinstance(s, slice) else str(s) for s in self]
        return '({})'.format(', '.join(plot_elem))

    def __repr__(self):
        """
        Get a string representation for the SampleIndex

        :return: the string representation
        :rtype: str
        """
        return '<{} {}>'.format(type(self).__name__, str(self))

    @property
    def isslice(self):
        """
        Flag indicating that the SampleIndex is a slice (as opposed to a
        simple single index).
        """
        return any(isinstance(x, slice) for x in self)

    def expand(self, size):
        """
        Function expanding a slice SampleIndex into a list of non-slice
        SampleIndex objects

        :param size: the size of the collection to slice
        """
        if len(size) != len(self):
            raise exceptions.FastrValueError('Number of dimensions of the index and size do not'
                                             ' match ({} vs {})'.format(len(self), len(size)))

        if not self.isslice:
            return self

        # For each dimension get the range arguments
        index_range = (x.indices(s) if isinstance(x, slice) else (x, x+1) for x, s in zip(self, size))

        # Create an xrange iterator for each dimension
        index_range = [xrange(*x) if len(x) > 0 else [] for x in index_range]

        # Create all combinations of indices
        return tuple(SampleIndex(x) for x in itertools.product(*index_range))


class SampleValue(MutableMapping):
    """
    A collection containing the content of a sample
    """
    _key_type = (int, str)

    def __init__(self, *args, **kwargs):
        try:
            self._data = dict(*args, **kwargs)
        except (TypeError, ValueError) as err:
            raise exceptions.FastrTypeError('Cannot create dict based on: args "{}" kwargs and "{}" (reason: {})'.format(args, kwargs, err))

    def __repr__(self):
        return '<SampleValue {}>'.format(self._data)

    def __getitem__(self, item):
        if not isinstance(item, self._key_type):
            raise exceptions.FastrTypeError('Keys in a SampleValue should be of {}'.format(self._key_type))

        try:
            return self._data[item]
        except KeyError:
            raise exceptions.FastrKeyError('Key {} not found in {}'.format(item, self._data))

    def __setitem__(self, key, value):
        if not isinstance(key, self._key_type):
            raise exceptions.FastrTypeError('Keys in a SampleValue should be of {}'.format(self._key_type))

        if not isinstance(value, tuple):
            raise exceptions.FastrTypeError('Values in a SampleValue should be of type tuple')

        self._data[key] = value

    def __getstate__(self):
        state = {}
        for key, value in self._data.items():
            state[key] = tuple(x.__getstate__() for x in value)

        return ('SampleValue', state)

    def __setstate__(self, state):
        if state[0] != 'SampleValue':
            raise exceptions.FastrValueError('Cannot unpickle incorrectly formatted SampleValue data!')
        self._data = {}
        for key, value in state[1].items():
            new_value = []
            for item in value:
                if item[0] not in fastr.typelist:
                    fastr.toollist.data  # Trigger toollist population
                new_item = fastr.typelist[item[0]]()
                new_item.__setstate__(item)
                new_value.append(new_item)

            self._data[key] = tuple(new_value)

    def __delitem__(self, key):
        if not isinstance(key, self._key_type):
            raise exceptions.FastrTypeError('Keys in a SampleValue should be of {}'.format(self._key_type))

        del self._data[key]

    def __len__(self):
        return sum(len(x) for x in self.values())

    def __iter__(self):
        for key in sorted(self._data.keys()):
            yield key

    @property
    def is_sequence(self):
        return all(isinstance(x, int) for x in self.keys())

    @property
    def is_mapping(self):
        return not self.is_sequence

    def sequence_part(self):
        return tuple(x for k, v in self.items() if isinstance(k, int) for x in v)

    def mapping_part(self):
        return OrderedDict([(k, v) for k, v in self.items() if isinstance(k, str)])

    def cast(self, datatype):
        new_data = {}
        for key, value in self.items():
            new_data[key] = tuple(x if datatype.isinstance(x) else datatype(x) for x in value)
        return SampleValue(new_data)

    def iterelements(self):
        for value in self.values():
            for element in value:
                yield element

    def __radd__(self, other):
        # Allow for sum
        if not isinstance(other, int):
            return NotImplemented

        if other != 0:
            return NotImplemented

        return self

    def __add__(self, other):
        if not isinstance(other, SampleValue):
            return NotImplemented

        # Create result with the tuple parts if they exist
        result = {}
        sequence_part_self = self.sequence_part()
        sequence_part_other = other.sequence_part()

        if len(sequence_part_self) > 0:
            result[0] = sequence_part_self

        if len(sequence_part_other) > 0:
            result[1] = sequence_part_other

        # Add keys known in self, merge with other if needed
        mapping_part_self = self.mapping_part()
        mapping_part_other = other.mapping_part()
        for key, value in mapping_part_self.items():
            if key in mapping_part_other:
                result[key] = value + mapping_part_other[key]
            else:
                result[key] = value

        # Add the keys of other not yet merged
        for key, value in mapping_part_other.items():
            if key not in result:
                result[key] = value

        return SampleValue(result)


class SampleCollection(MutableMapping):
    """
    The SampleCollections is a class that contains the data including a form
    of ordering. Each sample is reachable both by its SampleId and a
    SampleIndex. The object is sparse, so not all SampleId have to be defined
    allowing for non-rectangular data shapes.

    .. note::

        This object is meant to replace both the SampleIdList and the
        ValueStorage.
    """

    def __init__(self, dimnames, parent):
        """
        Createa a new SampleCollection
        """
        self._dimnames = dimnames
        self._ndims = len(dimnames)
        self._size = (0,) * self.ndims
        self._parent = parent
        self._id_lookup = dict()
        self._index_lookup = dict()

    def __repr__(self):
        return '<SampleCollection containing {} samples>'.format(self.size)

    def __contains__(self, item):
        """
        Check if an item is in the SampleCollection. The item can be a SampleId
        or SampleIndex. If the item is a slicing SampleIndex, then check if it
        would return any data (True) or no data (False)

        :param item: the item to check for
        :type item: SampleId, SampleIndex
        :return: flag indicating item is in the collections
        :rtype: bool
        """
        if not isinstance(item, (SampleIndex, SampleId)):
            try:
                item = SampleIndex(item)
            except (exceptions.FastrValueError, exceptions.FastrTypeError):
                try:
                    item = SampleId(item)
                except (exceptions.FastrValueError, exceptions.FastrTypeError):
                    pass

        if isinstance(item, SampleId):
            return item in self._id_lookup
        elif isinstance(item, SampleIndex):
            if not item.isslice:
                return item in self._index_lookup
            else:
                return any(x in self._index_lookup for x in item.expand(self.size))
        else:
            return False

    def __getitem__(self, item):
        """
        Retrieve (a) SampleItem(s) from the SampleCollection using the SampleId
        or SampleIndex. If the item is a tuple, it should be valid tuple for
        constructing either a SampleId or SampleIndex.

        :param item: the identifier of the item to retrieve
        :type item: SampleId, SampleIndex, or tuple
        :returns: the requested item
        :rtype: SampleItem
        :raises FastrTypeError: if the item parameter is of incorrect type
        :raises KeyError: if the item is not found
        """

        # If the item is not a SampleId or SampleIndex, try to salvage by
        # turning the item into a proper format
        if not isinstance(item, (SampleIndex, SampleId)):
            try:
                item = SampleIndex(item)
            except (exceptions.FastrValueError, exceptions.FastrTypeError):
                try:
                    item = SampleId(item)
                except (exceptions.FastrValueError, exceptions.FastrTypeError):
                    pass

        if isinstance(item, SampleId):
            try:
                return self._id_lookup[item]
            except KeyError:
                raise exceptions.FastrKeyError('Cannot find key {!r} in {}'.format(item, sorted(self._id_lookup.keys())))
        elif isinstance(item, SampleIndex):
            if not item.isslice:
                try:
                    return self._index_lookup[item]
                except KeyError:
                    if all(0 <= x < s for x, s in zip(item, self.size)):
                        raise exceptions.FastrIndexNonexistent('Could not find index {} in sparse array {}'.format(item, self))
                    else:
                        raise exceptions.FastrKeyError('Cannot find index {!r} in {}'.format(item, sorted(self._index_lookup.keys())))
            else:
                # Unpack slice
                try:
                    return tuple(self._index_lookup[x] for x in item.expand(self.size) if x in self._index_lookup)
                except KeyError:
                    raise exceptions.FastrKeyError('Cannot find key {!r}'.format(item))
        else:
            raise exceptions.FastrTypeError('The key of a SampleCollection has'
                                            'to be a SampleId or SampleIndex'
                                            ' (found {})'.format(item))

    def __setitem__(self, key, value):
        """
        Set an item to the SampleCollection. The key can be a SampleId,
        SampleIndex or a tuple containing a SampleId and SampleIndex.
        The value can be a SampleItem (with the SampleId and SampleIndex
        matching), a tuple with values (assuming no depending jobs), or a
        with a list of values and a set of depending jobs.

        :param key: the key of the item to store
        :type key: SampleId, SampleIndex, tuple of both, or SampleItem
        :param value: the value of the SampleItem to store
        :type value: SampleItem, tuple of values, or tuple of tuple of values and set of depending jobs
        :raises FastrTypeError: if the key or value types are incorrect
        :raises FastrValueError: if the id or values are incorrectly formed
        """

        # Parse the key to set
        if isinstance(key, SampleId):
            sample_id = key
            sample_index = None
        elif isinstance(key, SampleIndex):
            sample_id = None
            sample_index = key
        elif isinstance(key, SampleItem):
            sample_id = key.id
            sample_index = key.index
        elif isinstance(key, tuple):
            if len(key) == 2 and isinstance(key[0], SampleId) and isinstance(key[1], SampleIndex):
                sample_id = key[0]
                sample_index = key[1]
            elif len(key) == 2 and isinstance(key[0], SampleIndex) and isinstance(key[1], SampleId):
                sample_id = key[1]
                sample_index = key[0]
            else:
                raise exceptions.FastrValueError('If the key to set is a tuple, it must '
                                                 'contain exactly one SampleId and one SampleItem')
        else:
            raise exceptions.FastrTypeError('The key to set must be a SampleId, SampleIndex,'
                                            ' or tuple (found {})'.format(type(key).__name__))

        # Parse the value to set
        if isinstance(value, SampleItem):
            if sample_id is not None and value.id != sample_id:
                raise ValueError('The SampleId of the key and the value do not'
                                 ' correspond! ({} vs {})'.format(value.id, sample_id))

            if sample_index is not None and value.index != sample_index:
                raise ValueError('The SampleIndex of the key and the value do'
                                 ' not correspond! ({} vs {})'.format(value.index, sample_index))

            value = SampleItem(value.index, value.id, value.data, value.jobs, value.failed_annotations)
        elif isinstance(value, SamplePayload):
            if sample_id is not None and value.id != sample_id:
                raise ValueError('The SampleId of the key and the value do not'
                                 ' correspond! ({} vs {})'.format(value.id, sample_id))

            if sample_index is not None and value.index != sample_index:
                raise ValueError('The SampleIndex of the key and the value do'
                                 ' not correspond! ({} vs {})'.format(value.index, sample_index))
        elif isinstance(value, tuple):
            # Assume the (values, depending_jobs) value format
            if sample_index is None:
                raise exceptions.FastrValueError('Could not find a proper SampleIndex for item to store')

            if sample_id is None:
                raise exceptions.FastrValueError('Could not find a proper SampleId for item to store')

            if len(value) == 2 and isinstance(value[0], SampleValue) and isinstance(value[1], set):
                # Create the SampleItem
                value = SampleItem(sample_index, sample_id, value[0], value[1], set())
            if len(value) == 3 and isinstance(value[0], SampleValue) and isinstance(value[1], set) and isinstance(value[2], set):
                # Create the SampleItem
                value = SampleItem(sample_index, sample_id, value[0], value[1], set())
            else:
                raise exceptions.FastrValueError('The tuple value is not a (values, depending) or'
                                                 ' (values, depending, failed_annotations)'
                                                 ' jobs tuple or a tuple of values!')
        else:
            # Try to cast it SampleValue
            if not isinstance(value, SampleValue):
                value = SampleValue(value)

            # Assume a iterable of values
            # We know that datatype is callable
            # pylint: disable=not-callable
            value = SampleItem(sample_index, sample_id, value, set(), set())

        # Make sure we know both sample_index and sample_id
        if sample_index is None:
            sample_index = value.index
        if sample_id is None:
            sample_id = value.id

        # Check if items already exists and that both point to the same item
        if sample_index in self._index_lookup or sample_id in self._id_lookup:
            if self._index_lookup.get(sample_index) is not self._id_lookup.get(sample_id):
                raise exceptions.FastrValueError('The index and id of the item to set point to different items!')

        # Adapt the size
        self._size = tuple(max(x, y + 1) for x, y in zip(self._size, sample_index))

        # Store in both lookups
        self._index_lookup[sample_index] = value
        self._id_lookup[sample_id] = value

    def __delitem__(self, key):
        """
        Remove an item from the SampleCollection

        :param key: the key of the item to remove
        :type key: SampleId, SampleIndex, tuple of both, or SampleItem
        """
        if isinstance(key, SampleId):
            sample_id = key
            sample_index = self._id_lookup[sample_id].index
        elif isinstance(key, SampleIndex):
            sample_index = key
            sample_id = self._index_lookup[sample_index].id
        elif isinstance(key, SampleItem):
            sample_id = key.id
            sample_index = key.index
        elif isinstance(key, tuple):
            if len(key) == 2 and isinstance(key[0], SampleId) and isinstance(key[1], SampleIndex):
                sample_id = key[0]
                sample_index = key[1]
            elif len(key) == 2 and isinstance(key[0], SampleIndex) and isinstance(key[1], SampleId):
                sample_id = key[1]
                sample_index = key[2]
            else:
                raise exceptions.FastrValueError('If the key to set is a tuple, it must '
                                                 'contain exactly one SampleId and one SampleItem')
        else:
            raise exceptions.FastrTypeError('The key to remove must be a SampleId, SampleIndex,'
                                            ' or tuple (found {})'.format(type(key).__name__))

        # Check if items already exists and that both point to the same item
        if sample_index not in self._index_lookup or sample_id not in self._id_lookup:
            raise exceptions.FastrValueError('Either the SampleId or SampleIndex to remove was not found!'
                                             ' This indicates a data corruption int he SampleCollection!')

        if self._index_lookup.get(sample_index) is not self._id_lookup.get(sample_id):
            raise exceptions.FastrValueError('The index and id of the item to remove point to different items!')

        del self._index_lookup[sample_index]
        del self._index_lookup[sample_id]

    def __iter__(self):
        """
        Iterate over the indices
        """
        for index in sorted(self._index_lookup):
            yield index

    def __len__(self):
        """
        Get the number of samples in the SampleCollections.
        """
        return len(self._id_lookup)

    @property
    def dimnames(self):
        """
        The dimnames of the SampleCollection
        """
        return self._dimnames

    @property
    def ndims(self):
        """
        The number of dimensions in this SampleCollection
        """
        return self._ndims

    @property
    def parent(self):
        """
        The parent object holding the SampleCollection
        """
        return self._parent

    @property
    def size(self):
        """
        The size of the SampleCollection. The size is the largest index in
        every dimension. For a 2D SampleCollection with 2 entries (10, 2) and
        (6, 6) the size would be (10, 6). As that is the rectangular grid that
        contains all data points.
        """
        return self._size

    @property
    def fullid(self):
        """
        The full defining ID for the SampleIdList
        """
        if self.parent is not None:
            return '{}/SampleCollection'.format(self.parent.fullid)
        else:
            return 'UNKNOWN/SampleCollection'



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


def which(name):
    """
    Find executable by name on the PATH, returns the executable that will be
     found in case it is used for a Popen call
     """
    mode = os.X_OK
    exts = os.environ.get('PATHEXT', '').split(os.pathsep)

    for path in os.environ.get('PATH', '').split(os.pathsep):
        path = os.path.join(path, name)

        if os.access(path, mode):
            return path

        for ext in exts:
            pathext = path + ext

            if os.access(pathext, mode):
                return path

    return None


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import os
import subprocess
import sys
import traceback

import fastr
from fastr.core.baseplugin import PluginState
from fastr.execution.executionpluginmanager import ExecutionPlugin, JobAction
from fastr.execution.job import JobState
from fastr.utils.classproperty import classproperty
from fastr.utils.multiprocesswrapper import function_wrapper


def run_job(job, job_status):
    try:
        fastr.log.debug('Running job {}'.format(job.id))
        job_status[job.id] = JobState.running

        command = [sys.executable,
                   os.path.join(fastr.config.executionscript),
                   job.commandfile]

        with open(job.stdoutfile, 'a') as fh_stdout, open(job.stderrfile, 'a') as fh_stderr:
            proc = subprocess.Popen(command, stdout=fh_stdout, stderr=fh_stderr)
            proc.wait()

            fastr.log.debug('Subprocess finished')
        fastr.log.debug('Finished {}'.format(job.id))
    except Exception:
        exc_type, _, trace = sys.exc_info()
        exc_info = traceback.format_exc()
        trace = traceback.extract_tb(trace, 1)[0]
        fastr.log.error('Encountered exception ({}) during execution:\n{}'.format(exc_type.__name__, exc_info))
        job.info_store['errors'].append((exc_type.__name__, exc_info, trace[0], trace[1]))

    return job


class ProcessPoolExecution(ExecutionPlugin):
    """
    A local execution plugin that uses multiprocessing to create a pool of
    worker processes. This allows fastr to execute jobs in parallel with
    true concurrency. The number of workers can be specified in the fastr
    configuration, but the default amount is the ``number of cores - 1``
    with a minimum of ``1``.

    .. warning::

        The ProcessPoolExecution does not check memory requirements of jobs
        and running many workers might lead to memory starvation and thus an
        unresponsive system.
    """
    _status = (PluginState.uninitialized, 'Please use the test() function to check DRMAA capability')

    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None, nr_of_workers=None):
        super(ProcessPoolExecution, self).__init__(finished_callback, cancelled_callback, status_callback)

        self.pool = multiprocessing.Pool(processes=fastr.config.process_pool_worker_number)

    @classproperty
    def configuration_fields(cls):
        return {
            "process_pool_worker_number": (
                int,
                # Default number of workers is core - 1 (to assure system
                # responsiveness)
                max(multiprocessing.cpu_count() - 1, 1),
                "Number of workers to use in a process pool"
            )
        }

    def cleanup(self):
        super(ProcessPoolExecution, self).cleanup()

        # Close the multiprocess pool
        fastr.log.debug('Stopping ProcessPool')

        fastr.log.debug('Terminating worker processes...')
        self.pool.terminate()

        fastr.log.debug('Joining worker processes...')
        self.pool.join()

        fastr.log.debug('ProcessPool stopped!')

    @classmethod
    def test(cls):
        try:
            # See if we can make a Pool and then remove it
            fastr.log.debug('Creating Pool')
            pool = multiprocessing.Pool(processes=1)
            fastr.log.debug('Terminating Pool')
            pool.terminate()
            del pool
            _status = ('Loaded', '')
        except OSError:
            _status = ('Failed', 'Multiprocessing Failed ({}):\n{}'.format(sys.exc_info()[0].__name__,
                                                                           traceback.format_exc()))

        cls._status = _status

    def _job_finished(self, result):
        pass

    def _queue_job(self, job):
        # Check if the job is ready to run or must be held
        self.pool.apply_async(function_wrapper,
                              [os.path.abspath(__file__), 'run_job', job, self.job_status],
                              callback=self.job_finished)


if __name__ == '__main__':
    multiprocessing.freeze_support()
